from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from dataclasses import field
from html.parser import HTMLParser
import math
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Union
from typing import cast
import warnings

from alliance_platform.codegen.printer import TypescriptPrinter
from alliance_platform.codegen.printer import TypescriptSourceFileWriter
from alliance_platform.codegen.typescript import CallExpression
from alliance_platform.codegen.typescript import FunctionDeclaration
from alliance_platform.codegen.typescript import Identifier
from alliance_platform.codegen.typescript import ImportDefaultSpecifier
from alliance_platform.codegen.typescript import ImportSpecifier
from alliance_platform.codegen.typescript import JsxAttribute
from alliance_platform.codegen.typescript import JsxElement
from alliance_platform.codegen.typescript import MultiLineComment
from alliance_platform.codegen.typescript import PropertyAccessExpression
from alliance_platform.codegen.typescript import ReturnStatement
from alliance_platform.codegen.typescript import StringLiteral
from alliance_platform.codegen.typescript import UnconvertibleValueException
from alliance_platform.codegen.typescript import convert_to_node
from allianceutils.template import build_html_attrs
from allianceutils.template import is_static_expression
from allianceutils.template import parse_tag_arguments
from allianceutils.template import resolve
from allianceutils.util import underscore_to_camel
from django import template
from django.core.exceptions import ImproperlyConfigured
from django.forms.models import ModelChoiceIteratorValue
from django.template import Context
from django.template import Node
from django.template import NodeList
from django.template import Origin
from django.template import TemplateSyntaxError
from django.template.base import UNKNOWN_SOURCE
from django.template.base import FilterExpression
from django.template.base import TextNode
from django.utils.functional import LazyObject
from django.utils.html import format_html
from django.utils.safestring import SafeString
from django.utils.safestring import mark_safe

from ..bundler import get_bundler
from ..bundler.base import BaseBundler
from ..bundler.base import ResolveContext
from ..bundler.context import BundlerAsset
from ..bundler.context import BundlerAssetContext
from ..bundler.ssr import ImportDefinition
from ..bundler.ssr import SSRItem
from ..bundler.ssr import SSRSerializable
from ..bundler.ssr import SSRSerializerContext
from ..bundler.vite import ViteBundler
from ..forms.renderers import form_input_context_key
from ..prop_handlers import CodeGeneratorNode
from ..prop_handlers import ComponentProp
from ..settings import ap_frontend_settings
from ..util import transform_attribute_names

register = template.Library()

# Used to ensure that the import for renderComponent comes first in the generated code. This is so
# any config related items get loaded first.
HIGHEST_PRIORITY_IMPORT = 100


def resolve_prop(value: Any, node: ComponentNode, context: Context) -> ComponentProp | Any:
    """Resolve the prop class to use for the specified ``value``

    To add new handlers, add class to the list set in  ``settings.REACT_PROP_HANDLERS``
    """
    if isinstance(value, dict):
        return {k: resolve_prop(v, node, context) for k, v in value.items()}
    if isinstance(value, (list, tuple)):
        return list(resolve_prop(v, node, context) for v in value)
    if isinstance(value, ModelChoiceIteratorValue):
        return resolve_prop(value.value, node, context)  # type: ignore[attr-defined] # It has this value but no type info
    if isinstance(value, LazyObject):
        # unwrap lazy objects
        return value.__reduce__()[1][0]
    for handler in ap_frontend_settings.REACT_PROP_HANDLERS:
        if handler.should_apply(value, node, context):
            return handler(value, node, context)
    return value


@register.tag("component")
def component(parser: template.base.Parser, token: template.base.Token):
    """Render a React component with the specified props

    There are three ways to specify which component to render. The first is for a `"common component" <https://react.dev/reference/react-dom/components/common>`_
    which is to say a built-in browser component (e.g. ``div``)::

        {% component "h2" %}Heading{% endcomponent %}

    The other two are for using a component defined in an external file. These will be loaded via
    the specified bundler class (currently :class:`~alliance_platform.frontend.bundler.vite.ViteBundler`). With
    a single argument it specifies that the default export from the file is the component to use::

        {% component "components/Button" %}Click Me{% endcomponent %}

    With two arguments the first is the file path and the second is the named export from that file::

        {% component "components/Table" "Column" %}Name{% endcomponent %}

    The last option has a variation for using a property of the export. This is useful for components
    where related components are added as properties, e.g. ``Table`` and ``Table.Column``::

        {% component "components" "Table.Column" %}Name{% endcomponent %}

    Note that this is only available when using named exports; default exports don't support it due to
    ambiguity around whether the ``.`` indicates file extension or property access.

    You can omit the file extension - the above could resolve to ``components/Table.tsx`` (``.js`` and ``.ts`` are also
    supported). See :ref:`resolving_paths` for details on how the file path is resolved.

    Props are specified as keyword arguments to the tag::

        {% component "components/Button" variant="primary" %}Click Me{% endcomponent %}

    Additionally, a dict of props can be passed under the ``props`` kwarg::

        {% component "components/Button" variant="primary" props=button_props %}Click Me{% endcomponent %}

    Children can be passed between the opening ``{% component %}`` and closing ``{% endcomponent %}``. Whitespace
    is handled the same as in JSX:

    * Whitespace at the beginning and ending of lines is removed
    * Blank lines are removed
    * New lines adjacent to other components are removed
    * New lines in the middle of a string literal is replaced with a single space

    So the following are all equivalent::

        {% component "div" %}Hello World{% endcomponent %}

        {% component %}
            Hello world
        {% endcomponent %}


        {% component %}
            Hello
            world
        {% endcomponent %}


        {% component %}

            Hello world
        {% endcomponent %}

    Components can be nested::

        {% component "components/Button" type="primary" %}
            {% components "icons" "Menu" %}{% endcomponent %}
            Open Menu
        {% end_component %}

    You can use ``as <variable name>`` to store in a variable in context that can then be passed to another tag::

        {% component "icons" "Menu" as icon %}{% end_component %}
        {% component "components/Button" type="primary" icon=icon %}Open Menu{% end_component %}

    All props must be JSON serializable. :class:`~alliance_platform.frontend.prop_handlers.ComponentProp` can be used to define
    how to serialize data, with a matching implementation in ``propTransformers.tsx`` to de-serialize it.

    For example :class:`~alliance_platform.frontend.prop_handlers.DateProp` handles serializing a python ``datetime`` and
    un-serializing it as a native JS ``Date`` on the frontend. See :class:`~alliance_platform.frontend.prop_handlers.ComponentProp`
    for documentation about adding your own complex props.

    Components are rendered using the ``renderComponent`` function in ``ap_frontend_settings.REACT_RENDER_COMPONENT_FILE``. This can be modified as needed,
    for example if a new provider is required.

    .. note:: All props passed through are converted to camel case automatically (i.e. ``my_prop`` will become ``myProp``)

    Server Side Rendering (SSR)
    ---------------------------

    Components will automatically be rendered on the server. See :ref:`ssr` for details about how this works.

    To opt out of SSR pass ``ssr:disabled=True`` to the component after the component name::

        {% component 'components/Button.tsx' ssr:disabled=True %}...{% endcomponent %}

    Options
    -------

    Various options can be passed to the component tag. To differentiate from actual props to the component they are
    prefixed with `ssr:` for server side rendering options, `component:` for general component options, or `container:`
    for options relating to the container the component is rendered into.

    - ``ssr:disabled=True`` - if specified, no server side rendering will occur for this component
    - ``component:omit_if_empty=True`` - if specified, the component will not be rendered if it has no children. This is
      useful for when components may not be rendered based on permission checks
    - ``container:tag`` - the HTML tag to use for the container. Defaults to the custom element ``dj-component``.
    - ``container:<any other prop>`` - any other props will be passed to the container element. For example, to add
      an id to the container you can use ``container:id="my-id"``. Note that while you can pass a style string, it's
      likely to be of little use with the default container style ``display: contents``. Most of the time you can just
      do the styling on the component itself.

    For example::

        {% component 'core-ui' 'Button' ssr:disabled=True variant="Outlined"%}
            ...
        {% endcomponent %}
    """
    return parse_component_tag(parser, token)


class NestedComponentProp(ComponentProp):
    """
    This is used to describe a nested component rendered within the ``component`` tag::

        {% component "div" %}
            {% component "button" disabled %}
                Click Me
            {% endcomponent %}
        {% endcomponent %}

    See ``revivers.Component`` in ``ssrJsonRevivers.tsx`` for how this is handled in SSR.
    """

    def __init__(self, value: ComponentNode, node: ComponentNode, context: Context):
        super().__init__(value, node, context)
        self.component = value
        self.props = value.resolve_props(context)

    def __repr__(self):
        return f"NestedComponentProp({self.component.source}, {self.props})"

    def get_tag(self):
        return "Component"

    def get_representation(self, ssr_cache):
        return {
            # This value will be resolved by revivers.ComponentImport
            "component": self.component.source,
            "props": self.props,
        }

    def generate_code(self, generator: ComponentSourceCodeGenerator):
        return generator.create_jsx_element_node(self.component, self.props)


PropType = str | float | int | list["PropType"] | tuple["PropType"] | dict[str, "PropType"] | ComponentProp
PropsType = dict[str, PropType]


class ComponentProps(SSRSerializable):
    """Stores the props for a given component and handles serialization"""

    props: PropsType

    def __init__(self, props: PropsType):
        # Copy as ``add_prop`` modifies
        self.props = props.copy()
        # Remove form input context key if present - this can't be used beyond this point and exists only as
        # a workaround to pass extra context to widgets, see FormInputContextRenderer
        if form_input_context_key in self.props:  # type: ignore[comparison-overlap]
            self.props.pop(form_input_context_key)  # type: ignore[call-overload]

    def __repr__(self):
        return f"ComponentProps({self.props})"

    def get_tag(self):
        return "json"

    def _serialize_prop(self, value: PropType, ssr_context: SSRSerializerContext):
        if isinstance(value, dict):
            return {underscore_to_camel(k): self._serialize_prop(v, ssr_context) for k, v in value.items()}
        if isinstance(value, (list, tuple)):
            return [self._serialize_prop(v, ssr_context) for v in value]
        if isinstance(value, ComponentProp):
            return value.serialize(ssr_context)
        if isinstance(value, ComponentProps):
            return value.serialize(ssr_context)
        return value

    def serialize(self, ssr_context: SSRSerializerContext):
        """Serialize props to Dict that can then be JSON encoded

        Handles conversion of :class:`~alliance_platform.frontend.prop_handlers.ComponentProp` instances.

        Args:
            options: The options to use when serializing. In particular the options tell serialization how to handle
                resolving imports when dealing with nested components.
        """
        return self._serialize_prop(self.props, ssr_context)

    def has_prop(self, prop_name: str):
        return prop_name in self.props

    def add_prop(self, prop_name: str, value: PropType):
        """Add a new prop"""
        self.props[prop_name] = value

    def update(self, props: dict):
        self.props.update(props)

    def pop(self, prop_name: str, default_value):
        return self.props.pop(prop_name, default_value)


class ComponentSourceBase:
    """Base class used to identify what the source of a component is

    This is used to differentiate between components that need to be imported vs common
    components (e.g. div, button etc).
    """

    def as_tag(self):
        """Used in debugging. Should return the name of the component as it would be used in JSX"""
        raise NotImplementedError


@dataclass(frozen=True)
class CommonComponentSource(ComponentSourceBase, SSRSerializable):
    """Used for things that require no imports, e.g. standard DOM components

    See https://beta.reactjs.org/reference/react-dom/components/common for origin of
    terminology 'common component'.
    """

    name: str

    def serialize(self, ssr_context: SSRSerializerContext):
        return self.name

    def as_tag(self):
        return self.name


@dataclass(frozen=True)
class ImportComponentSource(ImportDefinition, ComponentSourceBase):
    """
    Used to identify a Component that needs to be imported from a module

    This differs from :class:`~alliance_platform.frontend.templatetags.react.CommonComponentSource` which is just
    a string (e.g. 'div' or 'button') and requires no import to work.
    """

    #: If specified, this is the name of the property to use from the import. e.g. "Table.Cell" would use the Cell property from Table.
    property_name: str | None = None

    def get_relative_path(self):
        return self.path.relative_to(get_bundler().root_dir)

    def get_tag(self):
        return "ComponentImport"

    def get_representation(self, context: SSRSerializerContext):
        return {
            "import": context.add_import(self),
            "propertyName": self.property_name,
        }

    def as_tag(self):
        if self.property_name:
            return f"{self.import_name}.{self.property_name}"
        return self.import_name


class NestedComponentPropAccumulator:
    """
    Used to accumulate nested components during render, and then applying the accumulated components to
    a final rendered value to return the children to use in a component props. This is easier to understand through an
    example (the redundant ``if`` is just to demonstrate a nested non-component tag)::

        {% component "div" %}
            {% if True %}{% component "strong" %}Test{% endcomponent %}{% endif %} Example
        {% endcomponent %}

    As a React element this would be::

        <div><strong>Test Component</strong></div>

    As django template nodes::

        ComponentNode(
            "div",
            [IfNode((Condition(True), ComponentNode("strong", [TextNode("Test")]), TextNode("Example")]
        )

    All the children of a node will be (by default) rendered to a string. We can handle direct ``ComponentNode``
    children by inspecting the type (we do exactly this in ``resolve_prop`` as an optimisation). For the general
    case where the ``ComponentNode`` could be anywhere down the tree we render as normal. ``ComponentNode.render``
    checks if it's within another component by calling ``NestedComponentPropAccumulator.get_current(context)``. If
    it is it calls ``add`` and returns the string. This will end up rendering something like::

        __NestedComponentPropAccumulator__prop__0 Example

    Internally ``NestedComponentPropAccumulator`` will store the actual prop::

        props = {
            "__NestedComponentPropAccumulator__prop__0": NestedComponentProp(
                ComponentNode("strong", ...),
                ComponentProps({"children": ["Test"]})
            )
        }

    Then when ``apply`` is called it will be passed the rendered string::

        "__NestedComponentPropAccumulator__prop__0 Example"

    And process that, returning a list that can be used as children for the component props::

        [
            NestedComponentProp(
                ComponentNode("strong", ...),
                ComponentProps({"children": ["Test"]})
            ),
            " Example"
        ]
    """

    #: Used internally to track where the current registry is stored in ``Context``
    context_key = "__NestedComponentPropAccumulator"
    #: The stored props as a mapping from the rendered placeholder string, to the ``NestedComponentProp``.
    props: dict[str, NestedComponentProp]
    #: The origin component node
    origin_node: ComponentNode

    @classmethod
    def get_current(cls, context: Context) -> NestedComponentPropAccumulator | None:
        """Get the current accumulator, if any

        This extracts the current accumulator instance from the template context. Returns ``None``
        if there is no active accumulator.
        """
        return context.get(cls.context_key, None)

    def __init__(self, context: Context, origin_node: ComponentNode):
        self.origin_node = origin_node
        self.context = context
        self.props = {}

    def __enter__(self):
        # push current context; matching pop is in ``__exit__``
        self.context.push()
        # store this instance in context
        self.context[self.context_key] = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.context.pop()

        # make sure apply was called
        if self.props:
            warnings.warn("NestedComponentPropAccumulator context exited without calling 'apply'")

    def add(self, prop: NestedComponentProp):
        """Add a prop to the accumulator

        This returns a string that should be rendered in the components place. It will then be
        handled in ``apply``.
        """
        key = f"{self.context_key}__prop__{len(self.props)}"
        self.props[key] = prop
        return key

    def apply(self, value: str):
        """Apply the accumulator props to ``value``.

        ``value`` will have placeholders for each prop that was accumulated. This method will
        return a list with each of those props in its correct position, surrounded by strings
        for all non-placeholder parts of ``value``.

        For example, given this value::

            "__NestedComponentPropAccumulator__prop__0 Example __NestedComponentPropAccumulator__prop__1 "

        This would be returned (details omitted)::

            [
                NestedComponentProp(...),
                " Example"
                NestedComponentProp(...),
            ]
        """

        # TODO: Not 100% sure about the behaviour below of stripping & excluding empty string - see how it goes.
        # There were definitely issues in _not_ doing this, but it's not clear if this is the right solution.
        children: list[NestedComponentProp | str] = []
        if self.props:
            prev_index = 0
            for placeholder, prop in self.props.items():
                index = value.find(placeholder)
                if index == -1:
                    warnings.warn(f"Unexpected: didn't find {placeholder} in string {value}")
                    continue
                if index > prev_index:
                    part = value[prev_index:index]
                    if part:
                        children.append(part)
                children.append(prop)
                prev_index = index + len(placeholder)
            value = value[prev_index:]
        if value:
            children += self.transform_string(value)
        self.props = {}

        return children

    def transform_string(self, child: str | NestedComponentProp):
        """Given a string, handle any necessary HTML conversions for React compatibility."""
        if isinstance(child, SafeString):
            return convert_html_string(
                child,
                self.origin_node.origin,
                wrap_new_node=lambda node: NestedComponentProp(node, self.origin_node, self.context),
            )
        return [child]


class ComponentSourceCodeGenerator:
    """Helper class to assist in generating source for a component tag

    This uses a ``TypescriptSourceFileWriter`` to generate the source code for a component tag but
    provides some additional helpers to make it easier to generate the source.

    The generated source code has two potential shapes depending on if ``requires_wrapper_component`` has been called.
    Without a wrapper component (in production the URLs would refer to built files)::

        import { TestComponent } from "http://localhost:5173/assets/frontend/src/components/TestComponent.tsx";
        import { renderComponent } from "http://localhost:5173/assets/frontend/src/renderComponent.tsx";

        renderComponent(
          document.querySelector("[data-djid='6126825472__1_3']"),
          TestComponent,
          {},
          "6126825472__1_3",
          true
        );

    With a wrapper component::

        import { TestComponent } from "http://localhost:5173/assets/frontend/src/components/TestComponent.tsx";
        import {
          createElement,
          renderComponent,
        } from "http://localhost:5173/assets/frontend/src/renderComponent.tsx";

        function Wrapper() {
          return createElement(TestComponent, {});
        }

        renderComponent(
          document.querySelector("[data-djid='6144913408__1_3']"),
          Wrapper,
          {},
          "6144913408__1_3",
          true
        );

    The second form is only required if one of the props needs to use a React hook (e.g. ``useViewModelCache``).
    """

    bundler: BaseBundler
    node: ComponentNode

    #: If a wrapper component is required. This is set by ``requires_wrapper_component``.
    _requires_wrapper_component: bool
    #: Tracks the name of all identifiers generated so far. This is used to ensure uniqueness.
    _used_identifiers: list[str]
    #: Used by ``create_jsx_element`` to detect when the template name changes so it can output a comment.
    _last_template_origin_name: str | None
    #: Used by ``create_jsx_element`` to track the specified value for the ``include_template_origin`` kwarg in the root node
    _last_include_template_origin: bool
    #: Any nodes that should be added to the start of the generated content. This can be used by ``ComponentProp`` handlers to add code outside the component.
    #: For example, populating a cache should be done once and not on every render.
    _leading_nodes: list[Node]

    def __init__(self, node: ComponentNode):
        self.node = node
        self.bundler = node.bundler
        self._writer = TypescriptSourceFileWriter(
            resolve_import_url=self._resolve_import_url,
        )
        self._requires_wrapper_component = False
        self._used_identifiers = []
        self._last_template_origin_name = None
        self._last_include_template_origin = False
        self._leading_nodes = []

        # Resolve the import for createElement and use the returned Identifier for the jsx_transform
        self._writer.jsx_transform = self._writer.resolve_import(
            ap_frontend_settings.REACT_RENDER_COMPONENT_FILE,
            ImportSpecifier("createElement"),
            import_order_priority=HIGHEST_PRIORITY_IMPORT,
        )

    def _resolve_import_url(self, path: Path | str):
        path = self.bundler.validate_path(path, resolve_extensions=[".ts", ".tsx"])
        return self.bundler.get_url(path)

    def resolve_component_import(self, component: ComponentNode):
        """Resolve import to use for a component."""
        if isinstance(component.source, ImportComponentSource):
            # If we are dealing with an import resolve it using TypescriptSourceFileWriter - it will
            # return the Identifier that can safely be used. The actual URL used is set above to
            # resolve using self.blunder.get_url
            identifier = self._writer.resolve_import(
                str(component.source.get_relative_path()),
                (
                    ImportDefaultSpecifier(component.source.import_name)
                    if component.source.is_default_import
                    else ImportSpecifier(component.source.import_name)
                ),
            )
            if component.source.property_name:
                return PropertyAccessExpression(identifier, Identifier(component.source.property_name))
            return identifier
        if isinstance(component.source, CommonComponentSource):
            # Nothing to import - just reference the component as a string
            # e.g. "div"
            return StringLiteral(component.source.name)
        raise NotImplementedError(f"No support for {component.source}")

    def resolve_prop_import(
        self, path: str | Path, specifier: ImportSpecifier | ImportDefaultSpecifier, import_order_priority=0
    ):
        """This is a separate method so that we can track prop imports separately.

        Any custom prop that needs and import should call this method

        This is necessary as the imports for props depend on usage and can't be statically determined. As such, these
        are tracked separated and added to the dynamic dependencies of the component. This allows the ``BundlerContext``
        to check this assets will be available in production and raise an error if not.
        """
        self.node.add_dynamic_dependency(self.bundler.validate_path(path, resolve_extensions=[".ts", ".tsx"]))
        return self._writer.resolve_import(path, specifier, import_order_priority=import_order_priority)

    def requires_wrapper_component(self):
        """Indicate that a wrapper component is required to render this component. This is required when using hooks.

        ``ComponentProp`` instances can call this method to indicate that a wrapper component is required.
        """
        self._requires_wrapper_component = True

    def _codegen_prop(self, value: PropType):
        if isinstance(value, dict):
            return {underscore_to_camel(k): self._codegen_prop(v) for k, v in value.items()}
        if isinstance(value, (list, tuple)):
            return [self._codegen_prop(v) for v in value]
        if isinstance(value, CodeGeneratorNode):
            return value.generate_code(self)
        if isinstance(value, ImportComponentSource):
            # This lets us pass through imports as props, for example to pass a component class itself as a prop to
            # another component
            return self.resolve_prop_import(
                value.path,
                (
                    ImportDefaultSpecifier(value.import_name)
                    if value.is_default_import
                    else ImportSpecifier(value.import_name)
                ),
            )
        return convert_to_node(value)

    def create_jsx_element_node(
        self, component: ComponentNode, resolved_props: ComponentProps, include_template_origin=None
    ):
        """
        Create a JSX element node for the specified component and props

        Note that this creates the representation of the component as a JSX element. The TypescriptPrinter will
        then convert it to code, either as JSX directly for debugging, or as React.createElement calls for code
        outputted to the browser (JSX can't be interpreted directly in the browser, but is very useful for debugging
        or test cases).

        Args:
            component: The component node to create the JSX element for
            resolved_props: The fully resolved props for the component
            include_template_origin: Whether to include a comment for each component indicating the template it was
                created from. Useful for debugging.

        Returns:
            A ``JsxElement`` that can be printed to code.
        """

        # If not specified, use the last value. This is useful for nested elements to use the value specified on the
        # root node.
        if include_template_origin is None:
            include_template_origin = self._last_include_template_origin
        last_include_template_origin = self._last_include_template_origin
        self._last_include_template_origin = include_template_origin

        # Track the template name so we know when it changes. This way we only output the template name when it changes.
        last_template_name = self._last_template_origin_name
        template_name = (
            component.origin.template_name.decode()
            if isinstance(component.origin.template_name, bytes)
            else component.origin.template_name
        )
        self._last_template_origin_name = template_name

        try:
            children = cast(
                list[NestedComponentProp | str] | str | NestedComponentProp,
                resolved_props.props.get("children", []),
            )
            attributes = [
                JsxAttribute(
                    StringLiteral(key.lower()) if "-" in key else Identifier(underscore_to_camel(key)),
                    self._codegen_prop(prop),
                )
                for key, prop in resolved_props.props.items()
                if key != "children"
            ]
            jsx_children = [
                self._codegen_prop(child)
                for child in ([children] if not isinstance(children, list) else children)
            ]
            leading_comments = None
            if include_template_origin and template_name and template_name != last_template_name:
                leading_comments = [MultiLineComment(template_name)]
            return JsxElement(
                self.resolve_component_import(component),
                attributes,
                jsx_children,
                leading_comments=leading_comments,
            )
        except UnconvertibleValueException as e:
            raise ValueError(
                f"Do not know how to handle prop of type {type(e.value)}: {e.value}\n\n "
                "Either add a handler to ALLIANCE_PLATEFORM['FRONTEND']['REACT_PROP_HANDLERS'] or check the correct value is being passed in."
            )
        finally:
            self._last_template_origin_name = last_template_name
            self._include_template_origin = last_include_template_origin

    def generate_code(self, props: ComponentProps, container_id: str):
        jsx_element = self.create_jsx_element_node(self.node, props)

        # This needs to happen after the jsx_element is created, as that is where props are processed that may trigger
        # leading nodes to be added
        for node in self._leading_nodes:
            self._writer.add_node(node)

        # In some cases a wrapper component is needed, e.g. when using props that require hooks. This just wraps
        # the element in a wrapper function and returns the element. The props themselves may be a hook call, so we
        # don't have to specifically check for hooks here.
        if self._requires_wrapper_component:
            wrapper_id = Identifier("Wrapper")
            self._writer.add_node(
                FunctionDeclaration(
                    wrapper_id,
                    [],
                    [ReturnStatement(jsx_element)],
                )
            )
            component_id = wrapper_id
            jsx_element = JsxElement(component_id, [], [])
        self._writer.add_node(
            CallExpression(
                self._writer.resolve_import(
                    ap_frontend_settings.REACT_RENDER_COMPONENT_FILE,
                    ImportSpecifier("renderComponent"),
                    import_order_priority=HIGHEST_PRIORITY_IMPORT,
                ),
                [
                    CallExpression(
                        PropertyAccessExpression(
                            Identifier("document"),
                            Identifier("querySelector"),
                        ),
                        [f"[data-djid='{container_id}']"],
                    ),
                    jsx_element,
                    container_id,
                    not self.node.ssr_disabled,
                ],
            )
        )
        return self._writer.get_code()

    def generate_identifier(self, name: str):
        """Generate a unique identifier for use in the generated code"""
        counter = 2
        original_name = name
        while name in self._used_identifiers:
            name = f"{original_name}{counter}"
            counter += 1
        self._used_identifiers.append(name)
        return Identifier(name)

    def add_leading_node(self, node: Node):
        """Add a node that should be added to the top of the generated code."""
        self._leading_nodes.append(node)


class OmitComponentFromRendering(Exception):
    """Raise this exception to indicate that the component should not be rendered.

    This can be used to indicate the component should not be rendered in certain circumstances, for example
    if a permission check is done to determine whether user can see certain content.
    """

    pass


class DeferredProp:
    """Represents a prop that should be deferred until the component is rendered.

    This is useful for cases where the prop is rendered outside the component, for example:

    .. code-block:: html

        {% url_with_perm "some-url" as url %}

        {% component "SomeComponent" url=url %}{% endcomponent %}

    In this case, the ``url`` prop is rendered outside the component and so needs to be deferred until the component
    is rendered so that things that need to happen in context work (e.g. raising ``OmitComponentFromRendering`` are caught).
    """

    def resolve(self, context: Context):
        raise NotImplementedError()


def _merge_strings(children: list[str | NestedComponentProp]) -> list[str | NestedComponentProp]:
    new_children: list[str | NestedComponentProp] = []
    current_str = ""
    for child in children:
        if isinstance(child, str):
            current_str += child
        else:
            if current_str:
                new_children += current_str.splitlines()
                current_str = ""
            new_children.append(child)
    if current_str:
        new_children += current_str.splitlines()
    return new_children


def process_component_children(children: list[str | NestedComponentProp]) -> list[str | NestedComponentProp]:
    """Process component children, reducing strings to match JSX behaviour

    This applies the following (from `React JSX docs <https://legacy.reactjs.org/docs/jsx-in-depth.html#string-literals-1>`_):

    * JSX removes whitespace at the beginning and ending of a line.
    * It also removes blank lines.
    * New lines adjacent to tags are removed;
    * New lines that occur in the middle of string literals are condensed into a single space.

    For our purposes the list contains either strings, or ``NestedComponentProp`` which indicates a 'tag' in the rules
    above.
    """
    processed_children: list[str | NestedComponentProp] = []
    current_str = ""

    # First combine adjacent strings
    children = _merge_strings(children)
    for i, item in enumerate(children):
        prev_is_str = isinstance(children[i - 1], str) if i - 1 >= 0 else None
        next_is_str = isinstance(children[i + 1], str) if i + 1 < len(children) else None
        if isinstance(item, str):
            if prev_is_str:
                item = item.lstrip()
            if next_is_str:
                item = item.rstrip()
            # Remove entirely blank lines
            if not item.strip():
                continue

            # Concatenate adjacent strings
            # This handled the condition: new lines that occur in the middle of string literals are condensed into a single space
            current_str += " " + item if current_str else item
        else:
            if current_str:
                processed_children.append(current_str)
                current_str = ""

            processed_children.append(item)
    if current_str:
        processed_children.append(current_str)
    return processed_children


# Used to identify children when processing them in resolve_props
class ChildrenList(list):
    pass


class ComponentNode(template.Node, BundlerAsset):
    """A template node used by :func:`~alliance_platform.frontend.templatetags.react.component`"""

    #: Any extra dependencies for this component. This comes from props used that may require imports, for example
    #: DateProp may require the date library be included.
    dynamic_dependencies: list[Path]

    container_tag: str | FilterExpression
    container_props: dict[str, Any]

    def __init__(
        self,
        origin: Origin,
        source: ComponentSourceBase,
        props: dict[str, Any],
        target_var=None,
        ssr_disabled: bool | FilterExpression = False,
        omit_if_empty: bool | FilterExpression = False,
        container_tag: str | FilterExpression = "dj-component",
        container_props: dict[str, Any] | None = None,
    ):
        if not ap_frontend_settings.REACT_RENDER_COMPONENT_FILE:
            raise ImproperlyConfigured(
                "When using the `react` tag you must have `REACT_RENDER_COMPONENT_FILE` set in ALLIANCE_PLATFORM['FRONTEND'] settings"
            )
        if "children" in props:
            props["children"] = ChildrenList(props["children"])
        self.container_tag = container_tag
        self.container_props = container_props or {}
        self.ssr_disabled = ssr_disabled
        self.source = source
        self.props = props
        self.target_var = target_var
        self.omit_if_empty = omit_if_empty
        # For the case where props=my_dict is passed
        self.extra_props = props.pop("props", None)
        self.dynamic_dependencies = []
        super().__init__(origin)

    def __repr__(self):
        return f"ComponentNode({self.source}, {self.props})"

    def add_dynamic_dependency(self, path: Path):
        """Track when a dynamic dependency is used.

        A dynamic dependency is JS file that can't be statically analyzed. For example, if a ``ComponentProp`` required
        an import (e.g. ``useViewModelCache`` for ``ViewModelProp``) then that would be a dynamic dependency. This lets
        us identify during dev imports that need to be explicitly included in the created bundle.
        """

        # If path is explicitly included in ``get_paths_for_bundling`` we don't have to worry about it as it
        # will always be included
        if path not in self.get_paths_for_bundling():
            self.dynamic_dependencies.append(path)

    def get_dynamic_paths_for_bundling(self) -> list[Path]:
        return self.dynamic_dependencies

    def get_paths_for_bundling(self) -> list[Path]:
        paths = [ap_frontend_settings.REACT_RENDER_COMPONENT_FILE]
        if isinstance(self.source, ImportComponentSource):
            paths.append(self.source.path)
        return paths

    def resolve_prop(self, value, context: Context):
        """Handles resolving values to a type that can be serialized

        If you add new :class:`~alliance_platform.frontend.prop_handlers.ComponentProp` there must a case here
        to convert values to the new type.
        """
        if isinstance(value, DeferredProp):
            value = value.resolve(context)
        if isinstance(value, (ChildrenList, NodeList)):
            children: list[str | NestedComponentProp] = []
            for child in value:
                if isinstance(child, ComponentNode):
                    # We could remove this branch - it's an optimisation of the below. We know the node type here
                    # directly so can avoid the extra work + string replacement that happens below.
                    try:
                        children.append(NestedComponentProp(child, self, context))
                    except OmitComponentFromRendering:
                        pass
                else:
                    with NestedComponentPropAccumulator(context, self) as accumulator:
                        # This will be a string but there may have been components that render (e.g. within
                        # other django tags like {% if %}, or from template inheritance and rendering into a
                        # block contained within a component).
                        child_value: str = child if isinstance(child, str) else child.render(context)
                        if child_value:
                            children += accumulator.apply(child_value)

            children = process_component_children(children)

            if self.omit_if_empty and not children:
                raise OmitComponentFromRendering()

            # Many things only expect a single child so handle that as a default. This isn't necessary as we handle
            # it in our calls to createElement, but makes for slightly more readable code so leaving it in.
            if len(children) == 1:
                return children[0]

            # NOTE: I removed this as we can handle it on the frontend by passing `children` through as a spread to
            # `React.createElement` which tells it the children are static. See ``ComponentSourceCodeGenerator.create_jsx_element_node``
            # for where this occurs. Adding keys here did cause some problems - namely with the `Cell` component in
            # `Table`; navigation with keyboard across rows broke.
            # for i, child in enumerate(children):
            #     if isinstance(child, required_imports) and not child.props.has_prop("key"):
            #         child.props.add_prop("key", i)

            return children
        # This won't be true for props that come from ``extra_props`` as it's already a dict before passed to the template
        # tag (the ``extra_props`` var itself is resolved in ``resolve_props``)
        if isinstance(value, FilterExpression):
            value = value.resolve(context)
            return self.resolve_prop(value, context)
        # Always handle this first as many things rely on NestedComponentProp being here
        if isinstance(value, ComponentNode):
            return NestedComponentProp(value, self, context)
        return resolve_prop(value, self, context)

    def resolve_props(self, context: Context) -> ComponentProps:
        """Resolve the props for this component to values that can be serialized

        To add special handling override the :meth:`~alliance_platform.frontend.templatetags.react.ComponentNode.resolve_prop`
        method.
        """
        props = self.props.copy()
        if self.extra_props:
            props.update(self.extra_props.resolve(context))
        return ComponentProps({key: self.resolve_prop(value, context) for key, value in props.items()})

    def render_component(self, context: Context):
        css_items = self.bundler.get_embed_items(self.get_paths_for_bundling(), "text/css")
        for item in css_items:
            self.bundler_asset_context.queue_embed_file(item)

        # This means this component is nested under another and needs to be handled by the parent
        # Add to the accumulator and return a string that will be handled by NestedComponentPropAccumulator.apply()
        accumulator = NestedComponentPropAccumulator.get_current(context)
        if accumulator:
            try:
                return accumulator.add(NestedComponentProp(self, self, context))
            except OmitComponentFromRendering:
                return ""

        if self.target_var:
            context[self.target_var] = NestedComponentProp(self, self, context)
            return ""

        props = self.resolve_props(context)
        asset_context = BundlerAssetContext.get_current()
        container_id = asset_context.generate_id()

        generator = ComponentSourceCodeGenerator(self)
        # Generate this first before queuing SSR so if it fails we don't queue the SSR item
        code = self.bundler.format_code(generator.generate_code(props, container_id).strip())
        if not self.ssr_disabled:
            ssr_placeholder = asset_context.queue_ssr(ComponentSSRItem(self.source, props, container_id))
        else:
            # Include a comment we can see in HTML output to assist with debugging
            ssr_placeholder = "<!-- SSR OPT OUT -->"
        if not asset_context.html_target.include_scripts:
            return ssr_placeholder

        html_attrs = build_html_attrs(
            {
                **{key: resolve(value, context) for key, value in self.container_props.items()},
                "data-djid": container_id,
            }
        )
        container_tag = resolve(self.container_tag, context)
        parts = [
            format_html(
                f"<{container_tag} {{}}>{{}}</{container_tag}>",
                html_attrs,
                mark_safe(ssr_placeholder),
            ),
            f'<script type="module">\n{code}\n</script>',
        ]
        if ap_frontend_settings.DEBUG_COMPONENT_OUTPUT:
            parts.append(f"<!--\n{self.print_debug_tree(props)}\n-->")
        return "\n".join(parts)

    def render(self, context: Context):
        try:
            return self.render_component(context)
        except OmitComponentFromRendering:
            return ""

    def print_debug_tree(self, props: ComponentProps, include_template_origin=True):
        """Print a debug tree of the component and its children.

        This renders to look like JSX with comments indicating which templates are used to render the component::

            { /* django/forms/widgets/select.html */ }
            <DjangoWidgetWrapper name="is_active">
              <Select >
                {/* django/forms/widgets/select_option.html */ }
                <Item key="unknown">
                  --------
                </Item>
              </Select>
            </DjangoWidgetWrapper>
        """
        # Disable jsx_transform so raw JSX is outputted rather than React.createElement calls
        printer = TypescriptPrinter(jsx_transform=None)
        generator = ComponentSourceCodeGenerator(self)
        jsx_element = generator.create_jsx_element_node(self, props, include_template_origin)

        return self.bundler.format_code(printer.print(jsx_element))


@register.simple_tag()
def react_refresh_preamble():
    """Add `react-refresh <https://www.npmjs.com/package/react-refresh>`_ support

    Currently only works with :class:`~alliance_platform.frontend.bundler.vite.ViteBundler`.

    This must appear after :meth:`~alliance_platform.frontend.templatetags.bundler.bundler_preamble`.

    See https://vitejs.dev/guide/backend-integration.html

    Usage::

        {% bundler_preamble %}
        {% react_refresh_preamble %}
    """
    bundler = get_bundler()
    if not isinstance(bundler, ViteBundler):
        raise ValueError("Current only ViteBundler is supported")
    if bundler.mode != "development":
        return ""
    return mark_safe(
        # Indentation level here chose to generate prettier HTML source ;)
        f"""
    <script type="module">
      import RefreshRuntime from '{bundler.get_url('@react-refresh')}';

      RefreshRuntime.injectIntoGlobalHook(window)
      window.$RefreshReg$ = () => {{}}
      window.$RefreshSig$ = () => (type) => type
      window.__vite_plugin_react_preamble_installed__ = true
    </script>
        """
    )


@dataclass
class ComponentSSRItem(SSRItem):
    """Represents a component that will be rendered on the server

    Provides payload containing necessary import & the props for rendering the component.

    See the ``ComponentSSRItem`` type in ``ssr.ts`` for where this is handled.

    NOTE: Keep in sync with ``ssr.ts``
    """

    #: Tells the renderer how to resolve the component source - either ``CommonComponentSource`` or ``ESMAssetSource``
    source: ComponentSourceBase
    #: Props for the component
    props: ComponentProps
    #: Used by React to prefix id's to guarantee uniqueness when multiple roots rendered on one page
    identifier_prefix: str

    def get_ssr_type(self):
        return "Component"

    def get_ssr_payload(self, ssr_context: SSRSerializerContext):
        """Matches ``ComponentSSRItem`` type in ``ssr.ts``


        For a common component this looks like::

            // {% component "div" %}Hello{% component %}
            {
              "component": "div",
              "props": {
                "children": ["Hello"]
              },
              "identifierPrefix": "abc123"
            }

        For an imported component source gets serialized as a custom format handled in ssrJsonRevivers.tsx ::

            // {% component "components/Input" value=5 %}{% endcomponent %}
            {
              "component": ["@@CUSTOM", "ComponentImport", {"import": "<generated string>", "propertyName": null}],
              "props": {
                "value": 5
              },
              "identifierPrefix": "6169587712__1_1"
            }

        After going through the reviver this becomes::

            {
                "component": [Function: Input],
                "props: { "value": 5 },
                "identifierPrefix": "6169587712__1_1"
            }
        """
        return {
            "component": self.source,
            "props": self.props,
            "identifierPrefix": self.identifier_prefix,
        }


@register.filter
def html_attr_to_jsx(attrs: dict):
    """Convert html attributes to casing expected by JSX

    Calls :meth:`~alliance_platform.frontend.util.transform_attribute_names`
    """
    return transform_attribute_names(attrs)


@register.filter
def merge_props(attrs1: dict, attrs2: dict | None):
    """Merge props from two dicts together

    Usage::

        {% component "MyComponent" props=widget.attrs|merge_props:some_more_props|html_attr_to_jsx %}{% endcomponent %}
    """
    if attrs2 is None:
        return attrs1
    new_props = {**attrs1, **attrs2}
    # TODO: merge class names. need to work out nomralization
    # ie. class vs class_name vs className
    return new_props


@register.filter
def none_as_nan(value):
    """Convert None to NaN

    This is useful for props that should be passed as NaN to the component. The NumberInput component uses `NaN`
    instead of `null` for no value.
    """
    if value is None:
        return math.nan
    return value


def get_component_source_from_tag_args(args: list[Any], tag_name: str, origin: Origin) -> ComponentSourceBase:
    if len(args) == 2:
        # This is the form
        #   {% component "@ant-design/icons" "MenuOutlined" %}
        # We treat this as a named import, e.g.
        #   import MenuOutlined from "@ant-design/icons"
        source = args[0]
        component_name = args[1]
        is_default_import = False
    else:
        # This is the form
        #   {% component "components/Button" %}
        # We treat this as a default import, e.g.
        #   import Button from "components/Button";
        # We could have a parameter to customise the name (Button here) but there's
        # no reason to as it's an internal detail of the codegen at that point.
        source = args[0]
        component_name = None
        is_default_import = True

    if not isinstance(source, str):
        raise TemplateSyntaxError(
            f"{tag_name} must be passed a string, received a {type(source)} witih value '{source}'"
        )
        # This is for react-dom HTML elements, e.g. "div" etc. This is identified by the tag
        # being alphabetic or one of the heading tags (these are the only html tags with numbers)
        # This doesn't support custom elements - if we need this we can extend this check to
        # consider a 'dash' as well.
        # Note that this means you can't import components from a lowercase file in the root,
        # e.g. {% component "mycomponent" %} won't work - it will consider that a common component.
        # Our conventions would disallow this anyway, and it will work so long as it's in a
        # subdirectory (e.g. "components/div" would work).
    if len(args) == 1 and (
        source.isalpha() and source.islower() or source in ["h1", "h2", "h3", "h4", "h5", "h6"]
    ):
        return CommonComponentSource(source)
    if not component_name:
        component_name = source.split("/")[-1].split(".")[0]
    bundler = get_bundler()
    resolver_context = ResolveContext(bundler.root_dir, origin.name)
    source_path = bundler.resolve_path(source, resolver_context, resolve_extensions=[".ts", ".tsx", ".js"])
    property_name = None
    if "." in component_name:
        parts = component_name.split(".")
        if len(parts) > 2:
            raise TemplateSyntaxError(f"Invalid component name {component_name}")
        component_name = parts[0]
        property_name = parts[1]
    return ImportComponentSource(source_path, component_name, is_default_import, property_name=property_name)


def parse_component_tag(
    parser: template.base.Parser,
    token: template.base.Token,
    *,
    node_class: type[ComponentNode] = ComponentNode,
    asset_source: ComponentSourceBase | None = None,
    container_tag: str = "dj-component",
    container_props: dict[str, Any] | None = None,
    props: dict[str, Any] | None = None,
    ssr_disabled: bool = False,
    omit_if_empty: bool = False,
    no_end_tag: bool = False,
):
    """Parse a component tag. Allows custom component tags to be created.

    Example usage::

        @register.tag("MyComponent")
        def my_custom_component(parser: template.base.Parser, token: template.base.Token):
            bundler = get_bundler()
            resolver_context = ResolveContext(bundler.root_dir, parser.origin.name if parser.origin else None)
            source_path = get_bundler().resolve_path(
                "path/to-component", resolver_context, resolve_extensions=[".ts", ".tsx", ".js"]
            )
            asset_source = ImportComponentSource(source_path, "MyComponentName", False)
            return parse_component_tag(parser, token, asset_source=asset_source)

    Args:
        parser: Parser as provided when using `@register.tag`
        token: Token as provided when using `@register.tag`
        node_class: The class to use for the node. This can be used to create custom nodes, should extend ``ComponentNode``
        asset_source: The source for the component. If not provided this will be determined from the tag arguments.
        container_tag: The tag to use for the container element. This defaults to ``dj-component``. This can also be
            specified when using the tag itself, e.g. ``{% MyComponent container:tag="div" %}``
        container_props: Any additional props to pass through to the container element
        props: Any default props to pass through to the component. These are combined with any props passed to the tag,
            with the tag props taking precedence.
        ssr_disabled: Whether to disable SSR for this tag. Can also be specified when using the tag, e.g. ``ssr:disabled=True``
        omit_if_empty: Whether to omit this tag if it has no children. Can also be specified when using the tag, e.g. ``component:omit_if_empty=True``

    Returns:
        The constructed ``ComponentNode``
    """
    tag_name = token.split_contents()[0]
    args, kwargs, target_var = parse_tag_arguments(
        parser,
        token,
        supports_as=True,
    )

    processed_args: list[str] = []
    context = Context()
    for arg in args:
        if not is_static_expression(arg):
            raise TemplateSyntaxError(
                f"{tag_name} must be passed static strings for its arguments (encountered variable '{arg.var}')"
            )
        else:
            processed_args.append(arg.resolve(context))
    args = processed_args
    if asset_source and args:
        raise TemplateSyntaxError(f"{tag_name} accepts no non-keyword arguments")
    elif not asset_source and len(args) not in [1, 2]:
        raise TemplateSyntaxError(f"{tag_name} tag requires either 1 or 2 arguments, received {len(args)}")
    if not no_end_tag:
        nodelist = parser.parse((f"end{tag_name}",))
        parser.delete_first_token()
    else:
        nodelist = None
    if "children" in kwargs:
        raise TemplateSyntaxError("Pass children as the tag contents rather than under the 'children' prop")
    namespaced_options: dict[str, dict[str, Any]] = defaultdict(dict)
    exclude_keys = []
    valid_keys = {
        "component": {"omit_if_empty"},
        "ssr": {"disabled"},
    }
    for k, v in kwargs.items():
        for namespace in ["container", "ssr", "component"]:
            prefix = namespace + ":"
            if k.startswith(prefix):
                key = k[len(prefix) :]
                if namespace in valid_keys and key not in valid_keys[namespace]:
                    raise TemplateSyntaxError(
                        f"Invalid option {k} passed to {tag_name}. Valid options are {' ,'.join(valid_keys[namespace])}"
                    )
                exclude_keys.append(k)
                namespaced_options[namespace][key] = v
    container_props = {**(container_props or {}), **namespaced_options["container"]}
    ssr_options = namespaced_options["ssr"]
    component_options = namespaced_options["component"]
    container_tag = container_props.pop("tag", container_tag)
    props = props.copy() if props else {}
    props.update({key: value for key, value in kwargs.items() if key not in exclude_keys})
    origin = parser.origin or Origin(UNKNOWN_SOURCE)
    if nodelist:
        # Convert the nodelist to a string with placeholders for the template nodes. This is then processed for HTML
        # contained within ``TextNode``s, before converting back to a list of nodes with any placeholders replaced with
        # the original template nodes. This allows us to handle HTML being passed to component children directly without
        # having to use __dangerouslySetInnerHTML.
        i = 0
        html = ""
        replacements: dict[str, Node] = {}
        for node in nodelist:
            if isinstance(node, TextNode):
                html += node.s
            else:
                placeholder = _html_replacement_placeholder_template.format(i)
                html += placeholder
                replacements[str(i)] = node
                i += 1
        props["children"] = convert_html_string(html, origin, replacements=replacements)
    if asset_source is None:
        asset_source = get_component_source_from_tag_args(args, tag_name, origin)

    return node_class(
        origin,
        asset_source,
        props,
        target_var=target_var,
        container_tag=container_tag,
        container_props=container_props,
        ssr_disabled=ssr_options.get("disabled", ssr_disabled),
        omit_if_empty=component_options.get("omit_if_empty", omit_if_empty),
    )


# These are used to replace template Nodes in a string with placeholders. The resulting string can them be parsed as HTML,
# and then the placeholders replaced with the original template nodes.
_html_replacement_placeholder_prefix = "_______PLACEHOLDER_____$"
_html_replacement_placeholder_suffix = "$_"
_html_replacement_placeholder_template = (
    _html_replacement_placeholder_prefix + "{0}" + _html_replacement_placeholder_suffix
)


@dataclass
class Element:
    name: str
    attrs: dict


@dataclass
class HTMLElement:
    tag: str
    attributes: dict[str, str] = field(default_factory=dict)
    children: list[Union["HTMLElement", str]] = field(default_factory=list)


class HtmlTreeParser(HTMLParser):
    def __init__(self):
        super().__init__(convert_charrefs=True)
        # The root of the parsed tree. The tag is nonsense, but not used - it's just to make implementation easier. We
        # just use it for ``children``
        self.root = HTMLElement("Root")
        # Stack to maintain hierarchy of elements
        self.stack = [self.root]

    def handle_starttag(self, tag, attrs):
        element = HTMLElement(tag, {attr[0]: attr[1] for attr in attrs})
        if self.stack:
            self.stack[-1].children.append(element)
        else:
            self.root = element
        self.stack.append(element)

    def handle_endtag(self, tag):
        if self.stack and self.stack[-1].tag == tag:
            self.stack.pop()

    def handle_data(self, data):
        if data and self.stack:
            self.stack[-1].children.append(data)


def convert_html_string(
    html: str,
    origin: Origin,
    *,
    replacements: dict[str, Node] | None = None,
    wrap_new_node: Callable[[ComponentNode], ComponentNode | ComponentProp] = lambda x: x,
):
    """
    Given a string that may contain HTML, convert it to a tree of ``ComponentNode``s

    Args:
        html: The html string to convert
        origin: The template origin
        replacements: Any replacements to make in the HTML after conversion. The way this works is that a template NodeList
            is processed into a string, with any template nodes replaced with placeholders. The string is converted to
            a component tree, then any placeholders are replaced with the original template nodes.
        wrap_new_node:
            Function to call to wrap any created ``ComponentNode``. This is used to wrap nodes in ``NestedComponentProp``
            when they are children of another component. This isn't necessary for top-level nodes as they are handled
            when ``parse_component_tag`` is first called.
    Returns:

    """
    # NOTE: I tried lxml initially (with & without BeautifulSoup), and it was slower for our specific use case.
    # In general, it's considered very fast, but we don't need most of its features and this simple parser was
    # faster.
    parser = HtmlTreeParser()
    # Parse the HTML
    parser.feed(html)

    tags = parser.root.children

    def handle_placeholders(content: str):
        if not replacements:
            return [content]
        parts = content.split(_html_replacement_placeholder_prefix)
        first = parts.pop(0)
        if not parts:
            return [first]
        children: list[str | Node] = []
        if first:
            children.append(first)
        for part in parts:
            placeholder_id, extra = part.split(_html_replacement_placeholder_suffix)
            children.append(replacements[placeholder_id])
            if extra:
                children.append(extra)
        return children

    def convert_attributes(attrs: dict[str, str | Any]):
        transformed = {}
        for key, value in attrs.items():
            parts = handle_placeholders(value)
            transformed[key] = parts[0] if len(parts) == 1 else NodeList(parts)
        return transformed

    def convert_tree(tree):
        children = []
        for el in tree:
            if isinstance(el, str):
                parts = handle_placeholders(el)
                children += parts
            else:
                children.append(
                    wrap_new_node(
                        ComponentNode(
                            origin,
                            CommonComponentSource(el.tag),
                            {**convert_attributes(el.attributes), "children": convert_tree(el.children)},
                        )
                    )
                )
        return children

    return convert_tree(tags)



from __future__ import annotations

from django.template import Origin
from django.template.base import UNKNOWN_SOURCE

from ..bundler import get_bundler
from ..bundler.base import ResolveContext
from ..templatetags.react import ImportComponentSource


def get_module_import_source(
    path: str, name: str, is_default_export: bool, origin: Origin | None, property_name: str | None = None
):
    """Create an ImportComponentSource for a node_modules module

    Args:
        name: The name of the export. This should be a named export from the alliance-ui package
        origin: The template origin
        property_name: Optional property name. Useful for exports like `Menubar` that have attached properties
            like `Menubar.Item`.

    Returns:
        The ``ImportComponentSource`` that can then be passed to ``parse_component_tag``
    """
    if origin is None:
        origin = Origin(UNKNOWN_SOURCE)
    bundler = get_bundler()
    resolver_context = ResolveContext(bundler.root_dir, origin.name)
    source_path = get_bundler().resolve_path(
        path, resolver_context, resolve_extensions=[".ts", ".tsx", ".js"]
    )
    return ImportComponentSource(source_path, name, is_default_export, property_name=property_name)



from __future__ import annotations

from django import template
from django.template import Library

from ..templatetags.react import parse_component_tag
from .utils import get_module_import_source


def menubar(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Menubar", False, parser.origin),
    )


def submenu(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source(
            "@alliancesoftware/ui", "Menubar", False, parser.origin, property_name="SubMenu"
        ),
    )


def menubar_item(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Item", False, parser.origin),
    )


def menubar_section(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Section", False, parser.origin),
    )


def register_menubar(register: Library):
    register.tag("Menubar")(menubar)
    register.tag("Menubar.SubMenu")(submenu)
    register.tag("Menubar.Item")(menubar_item)
    register.tag("Menubar.Section")(menubar_section)



from django import template
from django.template import Context
from django.template import Library
from django.utils.dateparse import parse_time

from ..templatetags.react import ComponentNode
from ..templatetags.react import ComponentProps
from ..templatetags.react import parse_component_tag
from .utils import get_module_import_source


class TimeInputNode(ComponentNode):
    def resolve_props(self, context: Context) -> ComponentProps:
        values = super().resolve_props(context)
        if values.has_prop("default_value") and isinstance(values.props["default_value"], str):
            values.update(
                {"default_value": self.resolve_prop(parse_time(values.props["default_value"]), context)}
            )
        return values


def time_input(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "TimeInput", False, parser.origin),
        node_class=TimeInputNode,
    )


def register_time_input(register: Library):
    register.tag("TimeInput")(time_input)



from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache
import json
from json import JSONDecodeError
import logging
import mimetypes
from pathlib import Path
import re
from typing import Callable
from typing import Iterable
from typing import cast
from urllib.parse import urljoin
import warnings

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.templatetags.static import static
import requests

from ..settings import ap_frontend_settings
from .base import AssetFileEmbed
from .base import BaseBundler
from .base import DevServerCheck
from .base import HtmlGenerationTarget
from .base import PathResolver

logger = logging.getLogger("alliance_platform.frontend")


def _create_html_tag(tag_name: str, attrs: dict[str, str]):
    """Helper to create an HTML tag"""
    is_void_element = tag_name in ["link", "img"]
    attrs_html = " ".join([f'{key}="{value}"' for key, value in attrs.items()])
    tag = f"<{tag_name} {attrs_html}".strip()
    if is_void_element:
        return f"{tag}>"
    return f"{tag}></{tag_name}>"


def get_content_type(src: str | Path | None) -> str | None:
    if not src:
        # If no source assume js, e.g. a common chunk
        return "text/javascript"
    ext = "".join(Path(src).suffixes).lstrip(".").lower()
    if ext in ["js", "jsx", "mjs", "ts", "tsx"]:
        return "text/javascript"
    if ext in ["css", "css.ts"]:
        return "text/css"
    return mimetypes.guess_type(src)[0]


class ViteManifestAssetMissingError(Exception):
    def __init__(self, manifest_file: Path, path: Path):
        super().__init__(f"{path} not found in manifest file '{manifest_file}'")


class AssetDependencies:
    #: Any dependencies not loaded dynamically. This includes any dependencies of dependencies.
    dependencies: list[ViteManifestAsset]
    #: Any dependencies loaded dynamically (i.e. by a ``import("asset")`` call). Includes any non-dynamic dependencies
    #: of the dependencies so that all requires assets can be preloaded if desired.
    dynamic_dependencies: list[ViteManifestAsset]

    def __init__(self, dependencies: list[ViteManifestAsset], dynamic_dependencies: list[ViteManifestAsset]):
        self.dependencies = dependencies
        self.dynamic_dependencies = dynamic_dependencies

    def get_js_dependencies(self) -> list[str]:
        """Returns list of javascript file names used"""
        deps = []
        for asset in self.dependencies:
            if asset.file not in deps:
                deps.append(asset.file)
        return deps

    def get_css_dependencies(self) -> list[str]:
        """Returns list of css file names used"""
        deps = []
        for asset in self.dependencies:
            for css_file in asset.css:
                if css_file not in deps:
                    deps.append(css_file)
        return deps

    def get_dynamic_js_dependencies(self) -> list[str]:
        """Returns list of javascript file names used by any dynamic imports"""
        deps = []
        for asset in self.dynamic_dependencies:
            if asset.file not in deps:
                deps.append(asset.file)
        return deps

    def get_dynamic_css_dependencies(self) -> list[str]:
        """Returns list of css file names used by any dynamic imports"""
        deps = []
        for asset in self.dynamic_dependencies:
            for css_file in asset.css:
                if css_file not in deps:
                    deps.append(css_file)
        return deps

    def merge(self, deps: AssetDependencies):
        """Merge two ``AssetDependencies`` together"""
        for dep in deps.dependencies:
            if dep not in self.dependencies:
                self.dependencies.append(dep)
        for dep in deps.dynamic_dependencies:
            if dep not in self.dynamic_dependencies:
                self.dynamic_dependencies.append(dep)


# Cache of collected dependencies. Done here instead of on class as the dataclass is frozen
collected_dependencies_cache: dict[ViteManifestAsset, AssetDependencies] = {}


@dataclass(frozen=True)
class ViteManifestAsset:
    """
    See https://github.com/vitejs/vite/blob/main/packages/vite/src/node/plugins/manifest.ts for what this looks like
    """

    #: The manifest this asset is referenced from
    manifest: ViteManifest
    #: The built file (e.g. a js or css file)
    file: str
    #: Whether this is an entry point. If ``False`` then the file hasn't been imported directly but is a dependency of another entry point.
    is_entry: bool
    #: Whether this is used as a dynamic entry point (i.e. imported using ``import()``
    is_dynamic_entry: bool
    #: The source file this was built from. For common chunks this won't be set.
    src: str | None
    #: Any CSS file dependencies
    css: tuple[str]
    #: Can exist in the manifest file but currently not used by this
    assets: tuple[str]
    #: Any other assets this asset imports
    imports: tuple[str]
    #: Any other assets this asset imports dynamically (i.e. using ``import()``
    dynamic_imports: tuple[str]

    def collect_dependencies(self) -> AssetDependencies:
        """
        Collect all dependencies for this asset, including itself.

        This is useful to resolve all the files that are needed for a particular asset.

        See ``AssetDependencies`` for more details
        """
        # This is done to avoid circular dependencies going into infinite recursion
        if self in collected_dependencies_cache:
            return collected_dependencies_cache[self]
        dependencies: list[ViteManifestAsset] = []
        dynamic_dependencies: list[ViteManifestAsset] = []
        # this asset is included in direct dependencies
        dependencies.append(self)
        asset_dependencies = AssetDependencies(dependencies, dynamic_dependencies)
        collected_dependencies_cache[self] = asset_dependencies
        for imp in self.imports:
            asset = self.manifest.get_asset(imp)
            if asset not in dependencies:
                dependencies.append(asset)
            asset_deps = asset.collect_dependencies()
            for _asset in asset_deps.dependencies:
                if _asset not in dependencies:
                    dependencies.append(_asset)
            for _asset in asset_deps.dynamic_dependencies:
                if _asset not in dynamic_dependencies:
                    dynamic_dependencies.append(_asset)
        for imp in self.dynamic_imports:
            asset = self.manifest.get_asset(imp)
            if asset not in dynamic_dependencies:
                dynamic_dependencies.append(asset)
            # Because `imp` is a dynamic import already all dependencies of it become dynamic
            asset_deps = asset.collect_dependencies()
            for _asset in asset_deps.dependencies + asset_deps.dynamic_dependencies:
                if _asset not in dynamic_dependencies:
                    dynamic_dependencies.append(_asset)
        for dep in dependencies:
            if dep in dynamic_dependencies:
                dynamic_dependencies.remove(dep)
        return asset_dependencies

    def get_content_type(self):
        """Get the content type for this asset

        For css & vanilla extract files this will be text/css, for typescript/js it will be
        text/javascript. For other files it will be guessed from the file extension (e.g.
        image/svg+xml for svg files or image/png for .png files).
        """
        return get_content_type(self.src)


class ViteManifest:
    """Contains entries from a Vite manifest file

    Vite generates a manifest.json file as part of build based on the ``build.manifest`` and ``build.ssrManifest``
    options. This files contains a mapping from each built file to the corresponding outputs (e.g. the built javascript
    file and css files).
    """

    #: The entries from the file indexed by the source path
    entries: dict[Path, ViteManifestAsset]
    #: The file path entries were read from
    manifest_file: Path
    #: The root that any absolute paths will be resolved relative to
    root_dir: Path

    def __init__(self, root_dir: Path, manifest_file: Path):
        """

        Args:
            root_dir: The root that any absolute paths will be resolved relative to
            manifest_file: The path to the manifest file. This will be read as JSON and used to populate ``entries``
        """
        self.root_dir = root_dir
        self.manifest_file = manifest_file
        if not manifest_file.exists():
            # We don't throw as this could happen when doing a build if `manifest.json` doesn't exist
            # yet, e.g. when calling `extract_frontend_assets`.
            warnings.warn(f"Manifest '{manifest_file} does not exist. Have you run `yarn build`?")
            return
        entries = {}
        for key, value in json.loads(manifest_file.read_text()).items():
            key = Path(key)
            entries[key] = ViteManifestAsset(
                manifest=self,
                file=value["file"],
                is_entry=value.get("isEntry", False),
                is_dynamic_entry=value.get("isDynamicEntry", False),
                # Note that this isn't necessarily the same as ``key`` although I think it is whenever it is set. For
                # common chunks that are extracted this won't be set (e.g. ``jsx-runtime-9393f670.js``).
                src=value.get("src", None),
                css=cast(tuple[str], tuple(value.get("css", []))),
                assets=cast(tuple[str], tuple(value.get("assets", []))),
                imports=cast(tuple[str], tuple(value.get("imports", []))),
                dynamic_imports=cast(tuple[str], tuple(value.get("dynamicImports", []))),
            )
            # for index files add a mapping for the directory as well, e.g. both of these:
            #  components/table/index.tsx
            #  components/table
            # will resolve to the same thing
            # This simplifies the implementation as in dev components/table will resolve but in the manifest it
            # will resolve to components/table/index.tsx
            if key.stem == "index":
                entries[key.parent] = entries[key]
        self.entries = entries
        for entry in self.entries.values():
            entry.collect_dependencies()

    def get_asset(self, path: Path | str):
        """Given ``path`` return the matching ``ViteManifestAsset`` or raise ``ViteManifestAssetMissingError`` if none found"""
        if not isinstance(path, Path):
            path = Path(path)
        if path.is_absolute():
            path = path.relative_to(self.root_dir)
        if path not in self.entries:
            raise ViteManifestAssetMissingError(self.manifest_file, path)
        return self.entries[path]


class ViteBundler(BaseBundler):
    """Bundler implementation for `Vite <https://vitejs.dev/>`_.

    Args:
        root_dir: The root path everything sits under; all other paths are resolved relative to this
        path_resolvers: A list of :class:`~alliance_platform.frontend.bundler.base.PathResolver` instances used to resolve paths
        server_build_dir: The directory SSR files are outputted to (see ``yarn build:ssr``)
        build_dir: The directory client side files are outputted to (see ``yarn build:client``)
        server_host: The hostname used for the dev server (e.g. ``127.0.0.1``)
        server_port: The port used for the dev server (e.g. ``5173``)
        server_protocol: The protocol used for the dev server (``http`` or ``https``)
        server_resolve_package_url: The URL, handled by the dev server, to use to resolve package requests in dev mode. This is
            only used for packages in /node_modules/ so that we can use the optimized versions of packages that Vite generates.
            Without this, you encounter errors like "Outdated Optimized Deps". The dev server should handle requests to this
            URL and redirect to the final package file. Example implementation for fastify, where ``server_resolve_package_url="redirect-package-url"``.

            .. code-block:: javascript

                server.get('/redirect-package-url/*', async (req, res) => {
                    const packagePath = (req.params as Record<string, string>)['*'];
                    const [, resolvedId] = await vite.moduleGraph.resolveUrl(packagePath);
                    // `resolveId` is an absolute path with query string. Resolve it relative to the project root
                    const relativePath = normalizePath(path.relative(projectDir, resolvedId));
                    // Convert the filesystem path to a web-accessible path
                    const url = path.join(vite.config.base, relativePath);
                    res.redirect(302, url);
                });
        mode: The mode the bundler is running in; one of ``development``, ``production`` or ``preview``
        wait_for_server: Function that can be passed that will be called before requesting assets for server. This function
            should handle waiting until the dev server is ready before returning (e.g. by polling the server status)
        production_ssr_url: URL for SSR (only used in production mode)

    """

    #: Directory server side rendering (SSR) files are compiled to
    server_build_dir: Path
    #: Directory client side files are compiled to
    build_dir: Path
    #: Manifest for SSR (only available when in production mode)
    server_build_manifest: ViteManifest
    #: Manifest for client build (only available when in production mode)
    build_manifest: ViteManifest
    #: The mode bundler is running in; either 'development', 'production' or 'preview'
    mode: str
    #: URL for SSR (only used in production mode)
    production_ssr_url: str | None
    #: Function that can be passed that will be called before requesting assets for server.
    wait_for_server: Callable[[], None] | None = None
    #: Path to the vite metdata JSON file that we use in dev to resolve optimised deps
    vite_metadata_path: Path

    def __init__(
        self,
        *,
        root_dir: Path,
        path_resolvers: list[PathResolver],
        server_build_dir: Path,
        build_dir: Path,
        server_host: str,
        server_port: str,
        server_protocol: str,
        server_resolve_package_url: str,
        mode: str,
        production_ssr_url: str | None = None,
        wait_for_server: Callable[[], None] | None = None,
    ):
        self.wait_for_server = wait_for_server
        valid_modes = ["development", "production", "preview"]
        if mode not in valid_modes:
            raise ValueError(f"'mode' must be one of {', '.join(valid_modes)}, received: '{mode}'")
        super().__init__(root_dir, path_resolvers)
        self.mode = mode
        self.server_build_dir = server_build_dir
        self.build_dir = build_dir
        self.node_modules_dir = ap_frontend_settings.NODE_MODULES_DIR
        if self.mode != "development":
            self.server_build_manifest = ViteManifest(self.root_dir, server_build_dir / "manifest.json")
            self.build_manifest = ViteManifest(self.root_dir, build_dir / "manifest.json")
        self.dev_server_url_base = ""
        self.dev_server_url = ""
        self.production_ssr_url = production_ssr_url
        if mode == "development":
            static_url = settings.STATIC_URL
            if not static_url:
                raise ImproperlyConfigured("Must set STATIC_URL if using Vite bundler in development mode")
            if not static_url.endswith("/"):
                static_url += "/"
            # TODO: KB29
            if static_url.startswith("http"):
                raise ValueError(f"static_url cannot be a full URL in {mode}")
            static_url = static_url.lstrip("/")
            # this is the base url for the dev server - use ``dev_server_url`` if referencing to assets
            self.dev_server_url_base = f"{server_protocol}://{server_host}:{server_port}"
            # this is what should be used anywhere assets need to be referred to
            self.dev_server_url = urljoin(self.dev_server_url_base, static_url)
            self.dev_server_resolve_package_url = urljoin(
                self.dev_server_url_base, server_resolve_package_url
            )
            if not self.dev_server_resolve_package_url.endswith("/"):
                self.dev_server_resolve_package_url += "/"
        elif mode == "preview":
            self.preview_url = f"{server_protocol}://{server_host}:{server_port}"

    def resolve_url(self, path: Path | str):
        """Resolve a URL to use to serve the specified ``path``

        In development this will be served from the dev server and in production from ``STATIC_URL``
        """
        if self.wait_for_server:
            self.wait_for_server()

        if not isinstance(path, Path):
            path = Path(path)
        if self.mode in ["production", "preview"]:
            return static(str(path))
        if path.is_absolute():
            path = path.relative_to(self.root_dir)
        return urljoin(self.dev_server_url, str(path))

    _vite_dev_metadata: dict | None = None
    _vite_dev_last_modified: float | None = None

    def get_url(self, path: Path | str):
        """Get the URL to load asset as ``path``

        In development this uses the dev server, in production is served from static files.
        """
        if self.mode == "development":
            # In dev, check the vite metadata to see if the file is in the optimized list - if so
            # load from the specified file. This resolved the need to explicitly include a bunch
            # of dependencies in vite.config.ts under optimizeDeps.include
            if str(path).startswith(str(self.node_modules_dir)):
                return urljoin(
                    self.dev_server_resolve_package_url, str(Path(path).relative_to(self.node_modules_dir))
                )
            return self.resolve_url(path)
        # production & preview both need to use the file from the manifest
        return self.resolve_url(self.build_manifest.get_asset(path).file)

    def does_asset_exist(self, filename: Path):
        """In production node_modules might not exist - instead check the manifest file"""
        if self.mode == "development":
            return super().does_asset_exist(filename)
        try:
            self.build_manifest.get_asset(filename)
            return True
        except ViteManifestAssetMissingError:
            return False

    def get_embed_items(
        self, paths: Path | Iterable[Path] | str, content_type: str | re.Pattern | None = None
    ):
        """
        Generate the necessary ``AssetFileEmbed`` instances for the specified asset(s) ``paths``.

        For example, a javascript file for a component could return an AssetFileEmbed instance for it's javascript content,
        and one for it's CSS content.

        Note that in development passing a '.css' file in ``paths`` will embed as a javascript file as that's how Vite
        works with HMR. See ``ViteEmbedCss`` for where this is handled.

        Note that if specific items require extra attributes on the tag (e.g. ``alt="my alt tag"``) then this can be
        attached to ``item.html_attrs``.

        TODO: Doing preload need to do testing in browser to work out best approach. In particular script modules - does
        rel="preload" work or do you have to use rel="modulepreload" (which has poor support)?

        Args:
            paths: The path(s) to the asset to embed. If you need to embed multiple assets it's best to do them all together
                so that any necessary de-duplication can occur.
            content_type: If set only assets of that type will be embedded, otherwise all asset will be. The two common content types
                are text/css and text/javascript, but other's like image/png are also possible.
        Returns:
            The list of ``AssetFileEmbed`` instances that will be embedded.
        """
        if isinstance(paths, str):
            paths = Path(paths)
        paths = [paths] if isinstance(paths, Path) else paths
        # don't use a set as we want to preserve ordering
        embed_items = []
        for path in paths:
            item = self._create_embed_item(path)
            if item not in embed_items and item.matches_content_type(content_type):
                embed_items.append(item)
            for dep_item in item.get_dependencies():
                if dep_item not in embed_items and dep_item.matches_content_type(content_type):
                    embed_items.append(dep_item)
        return embed_items

    def get_preamble_html(self):
        """In development returns HMR client setup for Vite

        NOTE: Does not include react-refresh, use ``{% react_refresh_preamble %}`` for that.
        """
        preload_tag = ""
        if self.mode == "development":
            preload_tag += _create_html_tag(
                "script", {"type": "module", "src": urljoin(self.dev_server_url, "@vite/client")}
            )
        return preload_tag

    def is_development(self):
        return self.mode == "development"

    @lru_cache()
    def resolve_ssr_import_path(self, path: Path | str) -> Path | str:
        """See :meth:`~alliance_platform.frontend.bundler.base.BaseBundler.resolve_ssr_import_path`"""
        if self.is_development():
            return path
        return self.server_build_dir / self.server_build_manifest.get_asset(path).file

    def get_development_ssr_url(self):
        """Returns the URL for SSR rendering used in dev

        This assumes the Vite dev server has a 'ssr/' endpoint to handle requests.
        """
        return urljoin(self.dev_server_url_base, "ssr")

    def get_ssr_url(self):
        """Returns the URL for SSR rendering

        In dev, this uses :meth:`~alliance_platform.frontend.bundler.vite.ViteBundler.get_development_ssr_url`.

        In production :data:`~alliance_platform.frontend.bundler.vite.ViteBundler.production_ssr_url` is used.
        """
        if self.is_development():
            return self.get_development_ssr_url()
        if self.mode == "preview":
            return urljoin(self.preview_url, "ssr")
        if self.production_ssr_url:
            return urljoin(self.production_ssr_url, "ssr")
        return None

    def get_ssr_headers(self):
        """In development add an X-SSR-ROOT-DIR header to the request to the dev server.

        This lets the dev server check the request came from the same project to avoid trying to SSR requests from
        a different project which results in confusing errors.
        """
        if self.is_development():
            return {"X-SSR-ROOT-DIR": str(self.root_dir)}
        return {}

    def check_dev_server(self):
        try:
            r = requests.get(urljoin(self.dev_server_url_base, "check"), timeout=1)
            if r.status_code != 200:
                return DevServerCheck(is_running=False)
            return DevServerCheck(is_running=True, project_dir=Path(r.json()["projectDir"]))
        except (requests.ConnectionError, requests.ConnectTimeout):
            return DevServerCheck(is_running=False)
        except requests.ReadTimeout:
            return DevServerCheck(is_running=True, read_timeout=True)

    def format_code(self, code: str):
        """In dev format code using /format-code endpoint defined in dev-server.ts

        In production this is a no-op for performance reasons.
        """
        if self.is_development() and (
            len(code) < ap_frontend_settings.DEV_CODE_FORMAT_LIMIT
            or ap_frontend_settings.DEV_CODE_FORMAT_LIMIT == 0
        ):
            if self.wait_for_server:
                self.wait_for_server()
            payload = {"code": code}
            try:
                response = requests.post(
                    urljoin(self.dev_server_url_base, "format-code"),
                    data=json.dumps(payload),
                    headers={"Content-Type": "application/json"},
                    timeout=ap_frontend_settings.DEV_CODE_FORMAT_TIMEOUT,
                )
                if response.status_code != 200:
                    logger.error(
                        f"Bad response {response.status_code} from dev-server.ts for format-code action: {response.content.decode()}"
                    )
                else:
                    try:
                        return response.json()["code"]
                    except JSONDecodeError:
                        logger.error(
                            f"Failed to decode JSON from code formatting, content received: {response.content.decode()}"
                        )
            except TypeError:
                logger.exception(f"Failed to encode JSON for code formatting. Payload was: {payload}")
            except requests.exceptions.Timeout:
                logger.error("Timed out connecting to vite server for code formatting")
            except requests.exceptions.ConnectionError:
                logger.error("Failed to connect to vite server for code formatting - is it running?")
            return code
        return code

    def _create_embed_item(self, path: Path) -> AssetFileEmbed:
        content_type = get_content_type(path)
        if not content_type:
            # If not specified assume it's javascript (e.g. `core-ui` would be `core-ui/index.tsx`)
            content_type = "text/javascript"
        args = [self, path, content_type]
        for ct, cls in class_by_content_type.items():
            if isinstance(ct, re.Pattern) and content_type:
                if ct.match(content_type):
                    return cls(*args)
            elif ct == content_type:
                return cls(*args)
        # Assume default is jS
        warnings.warn(f"Unknown content type {content_type} for {path}, assuming javascript")
        return ViteJavaScriptEmbed(self, path, content_type)


class ViteEmbed(AssetFileEmbed):
    """A embed item for Vite assets"""

    #: The bundler instance to use to resolve asset paths
    bundler: ViteBundler
    #: The path to the file to embed. This will be used to resolve the final URL to embed (e.g. bundled files in production, dev server URLs in dev)
    path: Path

    def __init__(
        self, bundler: ViteBundler, path: Path, content_type: str, html_attrs: dict[str, str] | None = None
    ):
        self.bundler = bundler
        self.path = path
        self.content_type = content_type
        super().__init__(html_attrs)

    def get_content_type(self) -> str:
        return self.content_type

    def __eq__(self, other):
        return self.path == other.path and self.bundler == other.bundler and type(self) is type(other)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.path})"


class ViteJavaScriptEmbed(ViteEmbed):
    """A JavaScript embed

    In development this loads from the dev server, in production it loads from the bundled file.
    """

    def get_dependencies(self) -> list[AssetFileEmbed]:
        if self.bundler.is_development():
            return []
        asset = self.bundler.build_manifest.get_asset(self.path)
        deps = asset.collect_dependencies()
        return [
            ViteCssEmbed(self.bundler, css_path, "text/css", is_resolved_from_manifest=True)
            for css_path in deps.get_css_dependencies()
        ]

    def generate_code(self, html_target: HtmlGenerationTarget):
        if not html_target.include_scripts:
            return ""
        if self.bundler.is_development():
            return _create_html_tag("script", {"src": self.bundler.resolve_url(self.path), "type": "module"})
        asset = self.bundler.build_manifest.get_asset(self.path)
        return _create_html_tag(
            "script", {**self.html_attrs, "src": self.bundler.resolve_url(asset.file), "type": "module"}
        )


class ViteCssEmbed(ViteEmbed):
    """A CSS embed

    In development this loads from the dev server, in production it loads from the bundled file.
    """

    #: If true, then ``path`` is the path to the CSS file in the manifest. If false, then ``path`` needs to be resolved from the manifest still.
    is_resolved_from_manifest: bool

    def __init__(self, *args, is_resolved_from_manifest: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_resolved_from_manifest = is_resolved_from_manifest

    def __eq__(self, other):
        return (
            self.path == other.path
            and self.bundler == other.bundler
            and type(self) is type(other)
            and self.is_resolved_from_manifest == other.is_resolved_from_manifest
        )

    def is_vanilla_extract_file(self):
        return self.path.name.lower().endswith(".css.ts")

    def get_dependencies(self) -> list[AssetFileEmbed]:
        # In dev we load CSS via JS, so that's handled in ``generate_code`` below. But in production the main
        # ``file`` in the manifest is a useful JS file for Vanilla Extract. So instead we return the
        # actual CSS files here (I assume it will always be a single file in practice but ``asset.css`` is a list).
        # For regular CSS files ``asset.css`` will be empty and ``asset.file`` will be the CSS file itself.
        if self.bundler.is_development():
            return []
        # In production return the CSS files specified in the manifest
        asset = self.bundler.build_manifest.get_asset(self.path)
        return [
            ViteCssEmbed(self.bundler, css_file, "text/css", is_resolved_from_manifest=True)
            for css_file in asset.css
        ]

    def generate_code(self, html_target: HtmlGenerationTarget):
        # In dev, we load CSS via JS as that's how Vite works (and makes it so hot reloading works)
        if self.bundler.is_development():
            if html_target.inline_css:
                # TODO: https://kanban.alliancesoftware.com.au/board/75/card/133649/
                warnings.warn(
                    "Inlining CSS is not currently supported in dev mode. CSS will be loaded via JS instead."
                )
            if self.is_vanilla_extract_file():
                from .vanilla_extract import resolve_vanilla_extract_class_mapping

                mapping = resolve_vanilla_extract_class_mapping(self.bundler, self.path)
                # In dev we load the styles via an intermediate TS file that imports the CSS file and setups
                # up hot module reloading. Without this intermediate any changes to CSS will cause a full page
                # reload rather than using HMR. See `vanillaExtractWithExtras.ts` for where this script is
                # created.
                fn = mapping.import_script_filename
                if fn is None:
                    warnings.warn(
                        f"Expected import_script_filename to be set for path {self.path}. Hot loading will not work for this file."
                    )
                else:
                    return _create_html_tag(
                        "script",
                        {
                            **self.html_attrs,
                            "src": self.bundler.resolve_url(fn),
                            "type": "module",
                            # This will block rendering until file loaded, which reduces flash of unstyled content a bit
                            # You will still get it for any components that loads it's own styles however.
                            "blocking": "render",
                        },
                    )
            return _create_html_tag(
                "script",
                {
                    **self.html_attrs,
                    "src": self.bundler.resolve_url(self.path),
                    "type": "module",
                },
            )
        if not self.is_resolved_from_manifest and self.is_vanilla_extract_file():
            # This is because the manifest contains a javascript file that does nothing when dealing with Vanilla Extract.
            # We don't want to load that, so we just ignore it. The actual CSS file is returned in ``get_dependencies``
            # above. This is a little complicated but necessary to load things correctly while handling dev where the JS
            # file is loaded, as well as handling both Vanilla Extract and regular CSS.
            # See https://kanban.alliancesoftware.com.au/board/75/card/133462/summary/
            return ""
        file = (
            self.path
            if self.is_resolved_from_manifest
            else self.bundler.build_manifest.get_asset(self.path).file
        )
        if html_target.inline_css:
            file = self.bundler.build_manifest.manifest_file.parent / file
            return f"<style>{file.read_text()}</style>"
        return _create_html_tag(
            "link",
            {**self.html_attrs, "rel": "stylesheet", "href": self.bundler.resolve_url(file)},
        )


class ViteImageEmbed(ViteEmbed):
    """An image embed

    This embeds as a ``img`` tag. In development this loads the image from the dev server, in production it loads from
    the bundled file. This is useful for cache busting (the bundled file just gets a hash in its name) or for using
    an image pipeline (see https://kanban.alliancesoftware.com.au/board/75/card/133647/ for possibility of this)
    """

    def can_embed_head(self):
        return False

    def generate_code(self, html_target: HtmlGenerationTarget):
        file = self.path
        if not self.bundler.is_development():
            asset = self.bundler.build_manifest.get_asset(self.path)
            if not len(asset.assets) == 1:
                warnings.warn(f"Expected 1 asset for image {self.path}, got {asset.assets}")
            if asset.assets:
                file = asset.assets[0]
        return _create_html_tag("img", {**self.html_attrs, "src": self.bundler.resolve_url(file)})


class_by_content_type = {
    "text/css": ViteCssEmbed,
    "text/javascript": ViteJavaScriptEmbed,
    re.compile(r"image/*"): ViteImageEmbed,
}



from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Pattern
from typing import TypedDict
from typing import Union

from alliance_platform.base_settings import AlliancePlatformSettingsBase
from django.utils.module_loading import import_string

if TYPE_CHECKING:
    from .bundler.asset_registry import FrontendAssetRegistry
    from .bundler.base import BaseBundler
    from .prop_handlers import ComponentProp


class AlliancePlatformFrontendSettingsType(TypedDict, total=False):
    """The type of the settings for the frontend of the Alliance Platform.

    You can set these in your django settings like::

        from pathlib import Path
        from typing import TypedDict
        from alliance_platform.codegen.settings import AlliancePlatformCodegenSettingsType
        from alliance_platform.core.settings import AlliancePlatformCoreSettingsType

        # What you define here depends on your setup
        PROJECT_DIR = Path(__file___).parent.parent

        # What you define here depends on which Alliance Platform modules you are using. At minimum, you need to include "CORE".
        class AlliancePlatformSettings(TypedDict):
            CORE: AlliancePlatformCoreSettingsType
            CODEGEN: AlliancePlatformCodegenSettingsType

        ALLIANCE_PLATFORM: AlliancePlatformSettings = {
            "CORE": {
                "PROJECT_DIR": PROJECT_DIR,
            },
            "FRONTEND": {
                "PRODUCTION_DIR": PROJECT_DIR / "frontend/build",
            }
        }

    Below are the valid keys for ``ALLIANCE_PLATFORM["FRONTEND"]``:
    """

    #: The bundler to use as either a string import path or the class instance itself
    BUNDLER: Union[str, "BaseBundler"]
    #: If true, the React template tag will include a more readable debug output in the HTML in a comment
    DEBUG_COMPONENT_OUTPUT: bool
    #: The asset registry for the bundler. This is used to add additional assets to the bundler that are not automatically discovered.
    #: This can be a string import path to the registry or the registry itself.
    FRONTEND_ASSET_REGISTRY: Union[str, "FrontendAssetRegistry"]
    #: A list of either ``re.Pattern`` or a :class:`~pathlib.Path`. If a template directory matches any entry it will be excluded from :class:`extract_frontend_assets <alliance_platform.frontend.management.commands.extract_frontend_assets.Command>`.
    #:
    #: If a :class:`~pathlib.Path` is used it will be checked if the directory starts with that path. Otherwise a ``re.Pattern`` will exclude a directory if it matches.
    EXTRACT_ASSETS_EXCLUDE_DIRS: tuple[Path | str, Pattern[str]]
    #: If set to a truthy value :func:`~alliance_platform.frontend.templatetags.bundler.bundler_dev_checks` will not display any HTML error, the error will only be available in the Django dev console.
    BUNDLER_DISABLE_DEV_CHECK_HTML: bool | None
    #: The path to the node_modules directory. This is used by ViteBundler to resolve optimized deps, and extract_frontend_assets to determine when an import comes from node_modules directly.
    NODE_MODULES_DIR: Path | str
    #: The directory production assets exists in. This directory should include the Vanilla Extract mappings.
    PRODUCTION_DIR: Path
    #: Any custom prop handlers to use for react components. This can be a string import path to a list of prop handlers, or the list directly.
    REACT_PROP_HANDLERS: str | list[type["ComponentProp"]]
    #: File that is used to render React components using the ``react`` tag. This file should export a function named ``renderComponent`` and a function ``createElement`` (this can just be re-exported from React).
    REACT_RENDER_COMPONENT_FILE: Path | str
    #: Set to a dotted path to a function that will be called to resolve the global context for SSR. This function should return a dictionary of values to be passed to the SSR renderer under the `globalContext` key.
    SSR_GLOBAL_CONTEXT_RESOLVER: str | None
    #: The limit to apply for code format requests in development mode. This is limited to 1mb by default; anything above that will not be formatted. This is only
    #: applicable to dev mode where code is formatted to make debugging easier when viewing the source.
    DEV_CODE_FORMAT_LIMIT: int | None
    #: This is the timeout to apply for code format requests in development mode. This is limited to 1 seconds by default. The only time you should need to
    #: tweak this is if you are attempting to debug issues with a large piece of code; in which case you likely need to increase ``DEV_CODE_FORMAT_LIMIT`` as well.
    DEV_CODE_FORMAT_TIMEOUT: int | None


def maybe_import_string(val: Any | None):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, str):
        return import_string(val)
    return val


class AlliancePlatformFrontendSettings(AlliancePlatformSettingsBase):
    """
    A settings object that allows alliance_platform.frontend settings to be accessed as
    properties. For example:

        from alliance_platform.frontend.settings import ap_frontend_settings
        print(ap_frontend_settings.PRODUCTION_DIR)

    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.

    This class is based off of the django-rest-framework settings module
    """

    #: The directory production assets exists in. This directory should include the Vanilla Extract mappings.
    PRODUCTION_DIR: Path
    #: The path to the node_modules directory. This is used by ViteBundler to resolve optimized deps, and extract_frontend_assets to determine when an import comes from node_modules directly.
    NODE_MODULES_DIR: Path
    #: Any custom prop handlers to use for react components
    REACT_PROP_HANDLERS: list[type["ComponentProp"]]
    #: If true, the React template tag will include a more readable debug output in the HTML in a comment
    DEBUG_COMPONENT_OUTPUT: bool
    #: The asset registry for the bundler. This is used to add additional assets to the bundler that are not automatically discovered.
    FRONTEND_ASSET_REGISTRY: "FrontendAssetRegistry"
    #: The bundler to use
    BUNDLER: "BaseBundler"
    #: Directories to exclude from asset extraction. By default, all directories returned by ``get_app_template_dirs("templates")`` will be inspected.
    EXTRACT_ASSETS_EXCLUDE_DIRS: tuple[Path | str, Pattern[str]]
    #: Disable the HTML outputted by ``bundler_dev_checks``. This can be useful if you want the check on generally in a project, but specific developers may wish to disable it.
    BUNDLER_DISABLE_DEV_CHECK_HTML: bool
    #: Set to a dotted path to a function that will be called to resolve the global context for SSR. This function should return a dictionary of values to be passed to the SSR renderer under the `globalContext` key.
    SSR_GLOBAL_CONTEXT_RESOLVER: str | None
    #: File that is used to render React components using the ``react`` tag. This file should export a function named ``renderComponent`` and a function ``createElement`` (this can just be re-exported from React).
    REACT_RENDER_COMPONENT_FILE: Path
    #: The limit to apply for code format requests in development mode. This is limited to 1mb by default; anything above that will not be formatted. This is only
    #: applicable to dev mode where code is formatted to make debugging easier when viewing the source.
    DEV_CODE_FORMAT_LIMIT: int
    #: This is the timeout to apply for code format requests in development mode. This is limited to 1 seconds by default. The only time you should need to
    #: tweak this is if you are attempting to debug issues with a large piece of code; in which case you likely need to increase ``DEV_CODE_FORMAT_LIMIT`` as well.
    DEV_CODE_FORMAT_TIMEOUT: int

    def check_settings(self):
        # TODO: Implement checks on required settings

        # lock registry to make sure assets aren't added after startup that would be missed by
        # extract_frontend_assets
        self.FRONTEND_ASSET_REGISTRY.lock()


IMPORT_STRINGS = [
    "REACT_PROP_HANDLERS",
    "BUNDLER",
    "FRONTEND_ASSET_REGISTRY",
]

DEFAULTS = {
    "REACT_PROP_HANDLERS": "alliance_platform.frontend.prop_handlers.default_prop_handlers",
    "DEBUG_COMPONENT_OUTPUT": False,
    "EXTRACT_ASSETS_EXCLUDE_DIRS": tuple(),
    "BUNDLER_DISABLE_DEV_CHECK_HTML": True,
    "SSR_GLOBAL_CONTEXT_RESOLVER": None,
    "REACT_RENDER_COMPONENT_FILE": None,
    "DEV_CODE_FORMAT_LIMIT": 1 * 1024 * 1024,
    "DEV_CODE_FORMAT_TIMEOUT": 1,
}


ap_frontend_settings = AlliancePlatformFrontendSettings(
    "FRONTEND", import_strings=IMPORT_STRINGS, defaults=DEFAULTS
)



from django.forms.renderers import TemplatesSetting


class form_input_context_key:
    pass


class FormInputContextRenderer(TemplatesSetting):
    """Renderer to accommodate passing extra context variables to widgets

    Django widget interface provides no way to pass extra context to widget
    templates short of extending the widget class, or adding it to widget.attrs.
    The problem with passing it in widget.attrs it that it's typically passed
    directly through to the widget. Polluting ``attrs`` with extra context is
    problematic both because it's unexpected and because you either just have to
    accept all the values and hope it causes no issues, or filter them out
    which is inconvenient in templates.

    This renderer works with ``form_input`` to pass extra context to widgets
    by setting a special key in the ``widget.attrs`` dictionary. The renderer
    then pops this value and adds the contents to the ``context`` that is then
    used to render the template.
    """

    form_input_context_key = form_input_context_key

    def render(self, template_name, context, request=None):
        if "widget" in context and "attrs" in context["widget"]:
            extra_context = context["widget"]["attrs"].pop(form_input_context_key, {})
            # setting a default means templates can assume the value always exists - makes
            # usage with merge_props etc easier
            if "core_ui_props" not in extra_context:
                extra_context["core_ui_props"] = {}
            context.update(extra_context)
        return super().render(template_name, context, request)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import re
from typing import Iterable

from django import template
from django.utils.safestring import SafeString


@dataclass
class ResolveContext:
    """Passed to :meth:`~alliance_platform.frontend.bundler.base.PathResolver.resolve` to project context for resolving paths"""

    #: The root dir for the project. Paths should sit under this.
    root_dir: Path
    #: The source path for the 'thing' path is being resolved for (e.g. a template file). May not be available.
    source_path: str | Path | None = None


class PathResolver:
    """A class that resolves a path string to a :class:`~pathlib.Path`

    This can be used to do things like resolve paths relative to a root directory, use path aliases etc.
    """

    def resolve(self, path: str, context: ResolveContext) -> Path | None:
        """Resolve a string ``path`` to a :class:`~pathlib.Path`.

        Should return ``None`` if this class was unable to resolve the path, in which case resolving will
        move onto the next resolver in order.
        """
        raise NotImplementedError


@dataclass
class RegExAliasResolver(PathResolver):
    """Resolves a path by replacing a substring using a regex"""

    #: The regular expression to match
    find: str
    #: The string to replace the match with
    replace: str

    def resolve(self, path: str, context: ResolveContext) -> Path | None:
        replaced_path = re.sub(self.find, self.replace, path)
        if replaced_path == path:
            return None
        return Path(replaced_path)


@dataclass
class SourceDirResolver(PathResolver):
    """Resolves any non-absolute path relative to ``base_dir``"""

    #: The directory to resolve paths relative to
    base_dir: Path

    def resolve(self, path: str, context: ResolveContext):
        if Path(path).is_absolute():
            return None
        return self.base_dir / path


class RelativePathResolver(PathResolver):
    """Resolves strings that begin with ``./`` or ``../`` as relative to ``context.source_path``"""

    def resolve(self, path: str, context: ResolveContext):
        if (path.startswith("./") or path.startswith("../")) and context.source_path:
            source = Path(context.source_path)
            if source.is_file():
                source = source.parent
            return (source / path).resolve()
        return None


@dataclass
class DevServerCheck:
    """Describes status of dev server"""

    #: True if dev server is running at the expected location (depends on the bundler)
    is_running: bool
    #: Read timeout
    read_timeout: bool = False
    #: The project dir the frontend dev server is running at. This is used to determine if it's for the same project as Django.
    project_dir: Path | None = None

    def is_wrong_server(self) -> bool:
        """Returns ``True`` if server is running, but it's for a different project"""
        from alliance_platform.frontend.bundler import get_bundler

        return self.is_running and get_bundler().root_dir != self.project_dir

    def is_ok(self) -> bool:
        """Returns ``True`` if server is good to use for this project"""
        return self.is_running and not self.is_wrong_server()


class BaseBundler:
    #: The root path everything sits under; all other paths are resolved relative to this
    root_dir: Path
    #: A list of :class:`~alliance_platform.frontend.bundler.base.PathResolver` instances used to resolve paths
    path_resolvers: list[PathResolver]

    def __init__(self, root_dir: Path, path_resolvers: list[PathResolver]):
        self.root_dir = root_dir
        self.path_resolvers = path_resolvers

    def is_development(self) -> bool:
        """Should return true if running in development mode"""
        raise NotImplementedError

    def get_url(self, path: Path | str):
        """Return the URL to load the specified asset at ``path``"""
        raise NotImplementedError

    def get_preamble_html(self) -> str:
        """Return preamble that is included in the HTML head

        Defaults to nothing. This can be used for things like setting up hot module reloading.

        Is used by the :meth:`~alliance_platform.frontend.templatetags.bundler.bundler_preamble` tag.
        """
        return ""

    def get_embed_items(
        self, paths: Path | Iterable[Path], content_type: str | re.Pattern | None = None
    ) -> list[AssetFileEmbed]:
        """
        Generate the necessary ``AssetFileEmbed`` instances for the specified asset(s) ``paths``.

        For example, a javascript file for a component could return an AssetFileEmbed instance for its javascript content
        plus one for it's CSS content.

        Args:
            paths: The path(s) to the files to embed. If you need to embed multiple assets it's best to do them all together
                so that any necessary de-duplication can occur.
            content_type: If set only assets of that type will be embedded, otherwise all asset will be. The two common content types
                are text/css and text/javascript, but other's like image/png are also possible.
        Returns:
            The list of ``AssetFileEmbed`` instances that will be embedded.
        """
        raise NotImplementedError()

    def resolve_path(
        self,
        path: str | Path,
        context: ResolveContext,
        suffix_whitelist: list[str] | None = None,
        suffix_hint: str | None = None,
        resolve_extensions: list[str] | None = None,
    ) -> Path:
        """Resolve a string to a :class:`~pathlib.Path` based on specified ``path_resolvers``

        This allows things like relative paths or resolving relative to a custom directory.

        Each resolver in :data:`~alliance_platform.frontend.bundler.base.BaseBundler.path_resolvers` is called
        in turn until one returns a :class:`~pathlib.Path` at which point it stops.

        Args:
            path: The path to resolve
            context: The context
            suffix_whitelist: (optional) Whitelist of suffixes to accept. Each suffix should include the leading '.'. If not specified
                    then all suffixes are supported.
            suffix_hint: (optional) Hint to display as part of error message if suffix not valid
            resolve_extensions: (optional) List of extensions to try if the path as specified doesn't exist.

        Returns:
            The resolved absolute :class:`~pathlib.Path`
        """
        resolved_path = None
        if isinstance(path, Path):
            path = str(path)
        if isinstance(path, SafeString):
            # force SafeString to string so things like ``Path`` work with it
            path = "" + path
        for resolver in self.path_resolvers:
            resolved_path = resolver.resolve(path, context)
            if resolved_path:
                break
        resolved_path = Path(resolved_path or path)
        if not resolved_path.is_absolute():
            resolved_path = self.root_dir / resolved_path
        return self.validate_path(
            resolved_path,
            suffix_whitelist=suffix_whitelist,
            suffix_hint=suffix_hint,
            resolve_extensions=resolve_extensions,
        )

    def validate_path(
        self,
        filename: str | Path,
        suffix_whitelist: list[str] | None = None,
        suffix_hint: str | None = None,
        resolve_extensions: list[str] | None = None,
    ) -> Path:
        """Validate a path exists, and optional has a suffix in the whitelist

        If ``resolve_extensions`` are specified the path will be checked with each extension in turn until
        one matches. If filename doesn't exist an error is raised.

        If ``suffix_whitelist`` is specified and the filename doesn't have a suffix in the whitelist an error is raised.

        Args:
            suffix_whitelist: (optional) Whitelist of suffixes to accept. Each suffix should include the leading '.'. If not specified
                    then all suffixes are supported.
            suffix_hint: (optional) Hint to display as part of error message if suffix not valid
            resolve_extensions: (optional) List of extensions to try if the path as specified doesn't exist.
        Returns:
            The resolved absolute :class:`~pathlib.Path` with any missing extensions added.
        """
        if not isinstance(filename, Path):
            filename = Path(filename)
        if not filename.is_absolute():
            filename = self.root_dir / filename
        if not filename.exists():
            attempted = []
            if resolve_extensions and not filename.suffixes:
                for ext in resolve_extensions:
                    p = filename.parent / (filename.name + ext)
                    if self.does_asset_exist(p):
                        filename = p
                        break
                    attempted.append(str(p.relative_to(self.root_dir)))
            if not self.does_asset_exist(filename):
                message = f"'{filename}' does not exist."
                if attempted:
                    message += f" Tried the following paths: {', '.join(attempted)}"
                if not filename.suffixes and not resolve_extensions:
                    message += " You must include the file extension unless referring to a directory with an index file."
                raise template.TemplateSyntaxError(message)

        if suffix_whitelist and "".join(filename.suffixes) not in suffix_whitelist:
            message = f"file must have one of the extension(s): {', '.join(suffix_whitelist)}. Received: '{filename}'."
            if suffix_hint:
                message = f"{message} {suffix_hint}"
            raise template.TemplateSyntaxError(message)
        return filename

    def does_asset_exist(self, filename: Path):
        """Check if asset exists. By default, looks to filesystem.

        Implementations may not rely on the filesystem, e.g. once a project is built and deployed
        the source may not exist and instead the built files are stored. In Vite, this would be checked
        by seeing if the asset exists in the manifest.
        """
        return filename.exists()

    def resolve_ssr_import_path(self, path: Path | str) -> Path | str:
        """
        Resolve a path for use in an ES import when server side rendering

        This is necessary as in dev importing from the file directly will work
        (e.g. ``import { Button } from 'components/Button'``) but in production
        it needs to be the built file (e.g. ``import { Button } from '/assets/Button-abc123.js'``).

        Note that this only applies for code that will be embedded directly in script tags - code
        that is bundled by the bundler will resolve this imports itself.
        Args:
            path: The path to transform into a value that can be used in an ES import

        Returns:
            A string that can be used in an ES import
        """
        raise NotImplementedError

    def get_ssr_url(self):
        """Return the URL to use to perform server side rendering

        :class:`~alliance_platform.frontend.bundler.middleware.BundlerAssetContextMiddleware` will ``POST`` to this URL
        in development to generate the HTML for any queued SSR items.

        Note that production SSR generation works with built files and so the bundler itself isn't responsible for
        handling requests, but it may provide a URL that we use (e.g. in PREVIEW mode with Vite).

        See :ref:`ssr` for details on how SSR works.
        """
        raise NotImplementedError

    def get_ssr_headers(self) -> dict[str, str]:
        """Return any headers to add to the SSR request"""
        return {}

    def check_dev_server(self) -> DevServerCheck:
        """Check if the dev server is running.

        Should return ``DevServerCheck`` which should indicate whether the server is running
        and if so, what the project dir for that server is. This allows detection of when
        a server is running, but it's for a different project.
        """
        raise NotImplementedError

    def format_code(self, code: str):
        """Form code for output. Typically, this would only occur in development for performance reasons.

        This exists on the bundler so that we can leverage existing dev server to avoid having to shell out to a prettier
        process each time.
        """
        return code


@dataclass
class HtmlGenerationTarget:
    """The generation target for HTML produced by templates rendered in :class:`~alliance_platform.frontend.bundler.context.BundlerAssetContext`

    This is intended to be used in conjunction with things like :class:`~alliance_platform.frontend.bundler.ssr.BundlerAssetServerSideRenderer`
    where the generated HTML may differ depending on how it will be used. For example while the browser may load
    CSS externally and run all scripts, using something like WeasyPrint to generate PDFs may not need to execute
    scripts at all and prefer to load CSS inline from <style> tags.
    """

    #: Label for the target for debugging purposes (e.g. "browser")
    label: str
    #: Whether this target should include scripts in the output
    include_scripts: bool
    #: Whether this target should output CSS inline in ``<style>`` tags.
    inline_css: bool


#: The default; for use with browser where scripts will be executed and CSS included from external files.
html_target_browser = HtmlGenerationTarget("browser", include_scripts=True, inline_css=False)
#: For PDFs; scripts will not be included (only used for SSR), CSS inlined in <style> tags
html_target_pdf = HtmlGenerationTarget("pdf", include_scripts=False, inline_css=True)


class AssetFileEmbed:
    """Represents a file that should be embedded in the HTML

    Each ``AssetFileEmbed`` is responsible for generating the HTML to embed the file. This can contain logic
    for doing things differently in dev vs production, or for different targets (e.g. browser vs PDF).

    These are created by the bundler when :meth:`~alliance_platform.frontend.bundler.base.BaseBundler.get_embed_items` is called.

    If extra HTML attributes are required for the tag, they can be set on the ``html_attrs`` dict.
    """

    #: Any HTML attributes to include in the tag
    html_attrs: dict[str, str]

    def __init__(self, html_attrs: dict[str, str] | None = None):
        self.html_attrs = html_attrs or {}

    def get_content_type(self) -> str:
        """Return the content type of the file to embed

        The most common types are text/css and text/javascript. Images will be image/<type>, e.g. image/png
        """
        raise NotImplementedError

    def get_dependencies(self) -> list[AssetFileEmbed]:
        """Return any additional dependencies for this file. For example a JS file may have some CSS associated with it."""
        return []

    def generate_code(self, html_target: HtmlGenerationTarget):
        """Generate the HTML code necessary to embed this file"""
        raise NotImplementedError

    def can_embed_head(self):
        """Should return ``True`` if this file can be embedded in HTML head rather than inline where used

        This is useful to load CSS before HTML renders to avoid un-styled content, and for javascript
        to start loading before it's used for performance

        Things that display inline cannot do this (e.g. images)
        """
        return True

    def matches_content_type(self, content_type: str | re.Pattern | None):
        """Return ``True`` if this file matches the given content type

        If ``content_type`` is ``None`` then it matches anything.
        """

        # if no content_type specified then we match anything
        if content_type is None:
            return True
        if isinstance(content_type, re.Pattern):
            return content_type.match(self.get_content_type())
        return self.get_content_type() == content_type



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
import warnings

from allianceutils.template import parse_tag_arguments
from allianceutils.template import resolve
from django import template
from django.forms import BaseForm
from django.forms import BoundField
from django.template import NodeList
from django.template import Origin
from django.template import TemplateSyntaxError
from django.template.base import UNKNOWN_SOURCE
from django.template.base import FilterExpression

from ..alliance_ui.misc import RawHtmlNode
from ..bundler import get_bundler
from ..bundler.base import ResolveContext
from ..forms.renderers import FormInputContextRenderer
from .react import ImportComponentSource

register = template.Library()


class _NOT_PROVIDED:
    pass


class FormInputNode(template.Node):
    origin: Origin
    show_valid_state: FilterExpression | bool
    help_text: FilterExpression | str | type[_NOT_PROVIDED]
    label: FilterExpression | str | type[_NOT_PROVIDED]
    is_required: FilterExpression | bool
    extra_attrs: dict[str, FilterExpression]

    def __init__(
        self,
        field: FilterExpression,
        origin: Origin | None,
        help_text=_NOT_PROVIDED,
        label=_NOT_PROVIDED,
        is_required=None,
        show_valid_state=True,
        **extra_attrs,
    ):
        self.origin = origin or Origin(UNKNOWN_SOURCE)
        self.show_valid_state = show_valid_state
        self.field = field
        self.help_text = help_text
        self.label = label
        self.required = is_required
        self.extra_attrs = extra_attrs or {}

    def render(self, context: template.Context):
        field: BoundField = self.field.resolve(context)
        form_context = FormContext.get_current(context)
        extra_attrs = {}
        if not isinstance(field.form.renderer, FormInputContextRenderer):
            warnings.warn("form_input tag should only be used with 'FormInputContextRenderer'")
        else:
            help_text = (
                resolve(self.help_text, context) if self.help_text is not _NOT_PROVIDED else field.help_text
            )
            if help_text:
                # Help text can be HTML and django docs make it clear this value is not HTML-escaped - so we
                # can safely embed it here as HTML using the RawHtmlNode
                bundler = get_bundler()
                resolver_context = ResolveContext(bundler.root_dir, self.origin.name)
                source_path = bundler.resolve_path(
                    "components/RawHtmlWrapper", resolver_context, resolve_extensions=[".ts", ".tsx", ".js"]
                )
                asset_source = ImportComponentSource(source_path, "RawHtmlWrapper", True)
                help_text = RawHtmlNode(
                    self.origin,
                    asset_source,
                    {"children": help_text},
                )
            extra_attrs[field.form.renderer.form_input_context_key] = {
                "raw_value": field.value(),
                "extra_widget_props": {
                    "label": resolve(self.label, context) if self.label is not _NOT_PROVIDED else field.label,
                    "errorMessage": ", ".join(str(e) for e in field.errors),
                    "validationState": (
                        "invalid"
                        if field.errors
                        else (
                            "valid"
                            if field.form.is_bound and resolve(self.show_valid_state, context)
                            else None
                        )
                    ),
                    "description": help_text,
                    "isRequired": (
                        resolve(self.required, context) if self.required is not None else field.field.required
                    ),
                    **{key: resolve(value, context) for key, value in self.extra_attrs.items()},
                    **form_context.get_extra_props(field),
                },
            }
            form_context.mark_processed(field.name)
        return field.as_widget(attrs=extra_attrs)  # type: ignore[arg-type] # dict type in form says on str|bool but other types seem to work fine


@register.tag("form_input")
def form_input(parser: template.base.Parser, token: template.base.Token):
    """Renders a form input with additional props supported by alliance-ui

    This sets `label`, `errorMessage`, `validationState`, `description` and `isRequired`. In addition, it may
    set `autoFocus` based on the `auto_focus` setting on the parent `form` tag.

    The following options can be passed to the tag to override defaults:

    - `label` - set the label for the input. If not specified will use ``field.label``.
    - `help_text` - help text to show below the input. If not specified will use ``field.help_text``.
    - `show_valid_state` - if true, will show the 'valid' (i.e. success) state of the input. If not specified will use
      ``False``. For most components in alliance-ui this results in it showing a tick icon and/or rendering green. If
      this is ``False`` only error states will be shown.
    - `is_required` - if true, will show the input as required. If not specified will use the model field ``required``
      setting.

    In addition, you can pass through any extra attributes that should be set on the input. For example, to set an
    addon for an alliance-ui ``TextInput`` you could do the following::

        {% form_input field addonBefore="$" %}

    Note that the attributes supported here depend entirely on the widget. If the widget is a React component, you
    can also pass react components to the tag::

        {% component "core-ui/icons" "Search" as search_icon %}${% endcomponent %}
        {% form_input field addonBefore=search_icon %}

    The additional props are added to the key ``extra_widget_props`` - so the relevant widget template needs to include
    this for the props to be passed through::

        {% component "@alliancesoftware/ui" "TextInput" props=widget.attrs|merge_props:extra_widget_props|html_attr_to_jsx type=widget.type name=widget.name default_value=widget.value %}
        {% endcomponent %}
    """
    tag_name = token.split_contents()[0]
    args, kwargs, target_var = parse_tag_arguments(parser, token, supports_as=True)
    if not len(args) == 1:
        raise TemplateSyntaxError(f"{tag_name} must be passed the form field to render input for")

    return FormInputNode(
        args[0],
        parser.origin,
        **kwargs,
    )


@dataclass
class FormContext:
    form: BaseForm
    #: If true, the first field with errors will be focused or the first field if form not submitted
    auto_focus: bool = False
    processed_fields: list[str] = field(default_factory=list)

    def mark_processed(self, field_name: str):
        self.processed_fields.append(field_name)

    @classmethod
    def get_current(cls, context: template.Context) -> FormContext:
        form_context = context.get("form_context", None)
        if not form_context:
            raise ValueError("form_input tag must be used within a form tag")
        return form_context

    _focused_field: BoundField | None = None

    def get_extra_props(self, field: BoundField):
        extra_props = {}
        if self.auto_focus:
            if self.form.is_bound and field.errors and not self._focused_field:
                self._focused_field = field
                extra_props["auto_focus"] = True
            elif not self.form.is_bound and not self.processed_fields:
                # focus first field
                extra_props["auto_focus"] = True
        return extra_props


class FormNode(template.Node):
    origin: Origin
    form: FilterExpression
    nodelist: NodeList
    auto_focus: FilterExpression | bool

    def __init__(self, form: FilterExpression, nodelist: NodeList, origin: Origin | None, auto_focus=False):
        self.origin = origin or Origin(UNKNOWN_SOURCE)
        self.form = form
        self.nodelist = nodelist
        self.auto_focus = auto_focus

    def render(self, context: template.Context):
        form: BaseForm = resolve(self.form, context)
        with context.push(form_context=FormContext(form, resolve(self.auto_focus, context))):
            return self.nodelist.render(context)


@register.tag("form")
def form(parser: template.base.Parser, token: template.base.Token):
    """Tag to setup a form context for form_input tags

    This tag doesn't render anything itself, it just sets up context for ``form_input`` tags. This is to support
    the ``auto_focus`` behaviour. This works by adding an ``auto_focus`` prop to the first field with errors, or the
    first rendered field if no errors are present.

    Usage::

        {% load form %}

        {% form form auto_focus=True %}
            <form method="post>
            {% for field in form.visible_fields %}
              {% form_input field %}
            {% endfor %}
            </form>
        {% endform %}


    .. note::

        Usage of this tag requires the following settings to be set::

            FORM_RENDERER = "alliance_platform.frontend.forms.renderers.FormInputContextRenderer"

    """
    tag_name = token.split_contents()[0]
    args, kwargs, target_var = parse_tag_arguments(parser, token, supports_as=True)
    if not len(args) == 1:
        raise TemplateSyntaxError(f"{tag_name} must be passed the django form instance")

    nodelist = parser.parse((f"end{tag_name}",))
    parser.delete_first_token()

    return FormNode(
        args[0],
        nodelist,
        parser.origin,
        **kwargs,
    )



from __future__ import annotations

from functools import lru_cache
from pathlib import Path
import threading
from typing import Type
import warnings

from django.conf import settings
from django.http import HttpRequest
from django.template import Origin
from django.utils.module_loading import import_string

from ..settings import ap_frontend_settings
from . import get_bundler
from .asset_registry import FrontendAssetRegistry
from .base import AssetFileEmbed
from .base import BaseBundler
from .base import HtmlGenerationTarget
from .base import html_target_browser
from .ssr import SSRItem

GLOBAL_BUNDLER_ASSET_CONTEXT = threading.local()


class NoActiveBundlerAssetContext(Exception):
    """Thrown if :meth:`~alliance_platform.frontend.bundler.context.BundlerAssetContext.get_current` is called when no context is active"""

    pass


class UndiscoverableAssetsError(Exception):
    """Thrown if assets are used, but cannot be discovered automatically and aren't registered with the asset registry"""

    def __init__(self, paths: set[Path]):
        paths_str = "\n".join(map(str, paths))
        super().__init__(
            f"The following paths were used but cannot be auto-discovered by `extract_frontend_assets`:\n{paths_str}\n\n"
            f"To resolve add these paths to the asset registry, e.g. `frontend_asset_registry.add_asset(...paths...)`."
        )
        self.undiscoverable_assets = paths


class BundlerAsset:
    """
    A class representing an asset (e.g. script, stylesheet) that needs to be bundled.

    Each asset could have dependencies that also need to be bundled. All relevant files should be returned by
    ``get_paths_for_bundling``.

    All assets must be static so that the files that are required in production can be extracted at build time. This allows
    the bundler to know what files to compile based on usage in templates. For example a Template node must be able to
    resolve its argument when the node is constructed - it cannot depend on resolving it based on context.

    On creation, each node is added to the current context with :meth:`~alliance_platform.frontend.bundler.context.BundlerAssetContext.add_asset`.

    The :class:`~alliance_platform.frontend.bundler.context.BundlerAssetContext` can then be queried to see what assets have been used.

    :class:`extract_frontend_assets <alliance_platform.frontend.management.commands.extract_frontend_assets.Command>` uses this to work out what assets are used anywhere in the app by loading all templates. Any
    template nodes that extend ``BundlerAsset`` will be automatically added to the context.
    """

    #: The current bundler
    bundler: BaseBundler
    #: The current context. Will always be set; if no context is active then an error will be thrown on init
    bundler_asset_context: BundlerAssetContext
    #: The template origin for this asset. This is used to ensure each asset will be found at build time.
    origin: Origin

    def __init__(self, origin: Origin):
        self.origin = origin
        self.bundler = get_bundler()
        # TODO: It's possible there's use cases where we might not want to require a context. For now require it until
        # they emerge.
        self.bundler_asset_context = BundlerAssetContext.get_current()
        self.bundler_asset_context.add_asset(self)

    def get_paths_for_bundling(self) -> list[Path]:
        """Return a list of paths to files this asset requires

        These will be included in the build process - see :class:`extract_frontend_assets <alliance_platform.frontend.management.commands.extract_frontend_assets.Command>`
        """
        raise NotImplementedError

    def get_dynamic_paths_for_bundling(self) -> list[Path]:
        """Return a list of paths to files this asset used at runtime.

        This is useful for cases where it can't be determined statically what dependencies are required (e.g. based
        on the usage of a component).

        This is used by ``BundlerAssetContext`` during development to determine that any used dependencies will be
        available in production.
        """
        return []


class AssetEmbedFileQueue:
    """Queue for asset files to be embedded in a page

    Preserves order file are added in & guarantees no duplicates will be added.
    """

    items: list[AssetFileEmbed]

    def __init__(self):
        self.items = []

    def add(self, item: AssetFileEmbed):
        if item not in self.items:
            self.items.append(item)

    def is_empty(self):
        return not self.items

    def clear(self):
        self.items = []


@lru_cache
def get_ssr_global_context_resolver():
    try:
        resolver = ap_frontend_settings.SSR_GLOBAL_CONTEXT_RESOLVER
    except AttributeError:
        return None
    else:
        return None if resolver is None else import_string(resolver)


class BundlerAssetContext:
    """
    Class that acts as container for global context data for React apps and a context manager

    When called in a django request and :class:`~alliance_platform.frontend.bundler.middleware.BundlerAssetContextMiddleware` is
    used a context always available from :meth:`~alliance_platform.frontend.bundler.context.BundlerAssetContext.get_current()`.

    Usage::

        with BundlerAssetContext() as context:
            # Any templates that contain nodes that extend BundlerAsset will end up in context
            contents = render_template()
            # Post process contents to handle SSR & embedding the asset tags
            contents = context.post_process(contents)

            # Retrieve the paths for all assets that were added to the context
            context.get_asset_paths()

    Usage in a django view when :class:`~alliance_platform.frontend.bundler.middleware.BundlerAssetContextMiddleware` is active::

        context = BundlerAssetContext.get_current()
        context.get_asset_paths()
    """

    #: All the assets that have been added
    assets: list[BundlerAsset]
    #: All assets that have been queued for embedding in the final HTML
    embed_item_queue: AssetEmbedFileQueue
    #: Contains all SSRItem's that have been queued for rendering
    ssr_queue: dict[str, SSRItem]
    #: The registry to use for assets that can't be discovered automatically.
    frontend_asset_registry: FrontendAssetRegistry
    #: The target HTML is being generated for. This determines things like whether scripts are required at runtime, and
    #: whether CSS should be inlined in ``style`` tags or linked from an external file. Defaults to
    html_target: HtmlGenerationTarget
    #: If specified, no checks will be done on context exit to ensure all assets have been embedded or post processed. Useful for tests.
    skip_checks: bool

    # Will only be set if ``bundler_embed_collected_assets`` is used
    embed_collected_assets_tag_placeholder: str | None

    #: Will be set if used within ``BundlerAssetContextMiddleware``. In other contexts may be unavailable
    request: HttpRequest | None

    def __init__(
        self,
        *,
        frontend_asset_registry: FrontendAssetRegistry | None = None,
        html_target=html_target_browser,
        skip_checks=False,
        request: HttpRequest | None = None,
    ):
        if frontend_asset_registry is None:
            frontend_asset_registry = ap_frontend_settings.FRONTEND_ASSET_REGISTRY
        self.request = request
        self.skip_checks = skip_checks
        self.html_target = html_target
        self.frontend_asset_registry = frontend_asset_registry
        self.assets = []
        self.ssr_queue = {}
        self.current_id = 0
        self.id_prefix = f"{threading.get_ident()}_{id(self)}_"
        self.is_middleware_registered = (
            "alliance_platform.frontend.bundler.middleware.BundlerAssetContextMiddleware"
            in settings.MIDDLEWARE
        )
        self.embed_item_queue = AssetEmbedFileQueue()
        self.embed_collected_assets_tag_placeholder = None

    def add_asset(self, asset: BundlerAsset):
        """Add an asset to the current context"""
        if asset not in self.assets:
            self.assets.append(asset)

    def queue_embed_file(self, item: AssetFileEmbed):
        """Queues an asset to be embedded in the final HTML

        The actual embedding and insertion into the HTML happens in :meth:`~alliance_platform.frontend.bundler.context.BundlerAssetContext.post_process`,
        which is called by :class:`~alliance_platform.frontend.bundler.middleware.BundlerAssetContextMiddleware` which
        relies on :meth:`~alliance_platform.frontend.templatetags.bundler.bundler_embed_collected_assets` being used in the template.
        Typically, this tag would exist in any base templates that are used for all pages.

        ``item`` is responsible for generating the actual HTML to embed which typically depends on the bundler.
        """
        self.embed_item_queue.add(item)

    def get_asset_paths(self, of_type: Type[BundlerAsset] | None = None) -> list[Path]:
        """Return the paths for assets used by assets added to this context"""
        asset_paths = []
        for asset in self.get_assets(of_type):
            for path in asset.get_paths_for_bundling():
                # we do it like this rather than Set so order is preserved
                if path not in asset_paths:
                    asset_paths.append(path)
        return asset_paths

    def get_assets(self, of_type: Type[BundlerAsset] | None = None) -> list[BundlerAsset]:
        """Get assets that have been added to context, optionally filtered by ``of_type``

        Args:
            of_type: If provided only assets of this type will be returned
        """
        if of_type:
            return [asset for asset in self.assets if isinstance(asset, of_type)]
        return self.assets

    @classmethod
    def get_current(cls) -> BundlerAssetContext:
        """Get the current BundlerAssetContext or raise :class:`~alliance_platform.frontend.bundler.context.NoActiveBundlerAssetContext` if none

        When called in a django request and :class:`~alliance_platform.frontend.bundler.middleware.BundlerAssetContextMiddleware` is
        used one will always be available.
        """
        if not hasattr(GLOBAL_BUNDLER_ASSET_CONTEXT, "contexts") or not GLOBAL_BUNDLER_ASSET_CONTEXT.contexts:
            raise NoActiveBundlerAssetContext("No context currently active")
        return GLOBAL_BUNDLER_ASSET_CONTEXT.contexts[-1]

    def queue_ssr(self, item: SSRItem) -> str:
        """Queue an item for server side rendering

        Args:
            item: The item to render

        Returns:
            A string which is the placeholder that should be outputted in the HTML. It will be replaced by
            :class:`~alliance_platform.frontend.bundler.middleware.BundlerAssetContextMiddleware`.
        """
        if not self.is_middleware_registered:
            raise ValueError(
                "`queue_ssr` cannot be used without `BundlerAssetContextMiddleware`. Add 'alliance_platform.frontend.bundler.middleware.BundlerAssetContextMiddleware' to `MIDDLEWARE`."
            )
        placeholder = f"<!-- ___SSR_PLACEHOLDER_{len(self.ssr_queue)}___ -->"
        self.ssr_queue[placeholder] = item
        return placeholder

    def generate_id(self) -> str:
        """Generate an ID that is guaranteed to be unique for this context

        This can be used for generating ID's to use in HTML, e.g. for container rendering React components into
        """
        next_id = f"{self.id_prefix}{self.current_id}"
        self.current_id += 1
        return next_id

    def __enter__(self):
        if not hasattr(GLOBAL_BUNDLER_ASSET_CONTEXT, "contexts"):
            GLOBAL_BUNDLER_ASSET_CONTEXT.contexts = []
        GLOBAL_BUNDLER_ASSET_CONTEXT.contexts.append(self)
        self.id_prefix = f"{self.id_prefix}_{len(GLOBAL_BUNDLER_ASSET_CONTEXT.contexts)}_"
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        GLOBAL_BUNDLER_ASSET_CONTEXT.contexts.pop()
        if exc_type or self.skip_checks:
            # Don't do any checks if there was an error
            return

        # Check all assets will be included in the production build
        unknown_assets = set()

        from ..management.commands.extract_frontend_assets import get_all_templates_files

        known_templates = get_all_templates_files()

        cleared_cache = False
        for asset in self.assets:
            origin_name = Path(asset.origin.name)
            if not asset.origin.template_name or origin_name not in known_templates:
                # handle case where new template has been added without server restart
                if not cleared_cache and asset.origin.template_name:
                    get_all_templates_files.cache_clear()
                    known_templates = get_all_templates_files()
                    cleared_cache = True
                    if Path(asset.origin.name) in known_templates:
                        continue
                unknown_assets.update(
                    self.frontend_asset_registry.get_unknown(*asset.get_paths_for_bundling())
                )
            unknown_assets.update(
                self.frontend_asset_registry.get_unknown(*asset.get_dynamic_paths_for_bundling())
            )
        if unknown_assets:
            raise UndiscoverableAssetsError(unknown_assets)

        if self.requires_post_processing():
            raise ValueError(f"{exc_type} BundlerAssetContext.post_process() was not called but is required")

    def register_embed_collected_assets_tag(self) -> str:
        """For use by ``bundler_embed_collected_assets`` only

        This is used by the template tag to register that it has been used, and to get the placeholder string to
        write in the initial HTML that will then be replaced in ``post_process``.
        """
        if self.embed_collected_assets_tag_placeholder:
            raise ValueError(
                "Duplicate {% bundler_embed_collected_assets %} tags detected. Only one such tag should exist."
            )
        # this is the string the template tag should render. It's then replaced in ``post_process``
        self.embed_collected_assets_tag_placeholder = "__BundlerAssetContext__embed_placeholder__"
        return self.embed_collected_assets_tag_placeholder

    def requires_post_processing(self):
        """Check if ``post_process`` needs to be called"""
        return not self.embed_item_queue.is_empty() or self.ssr_queue

    def abort_post_process(self):
        """Called when ``post_process`` can't be called (e.g. on a request where content-type isn't text/html)"""
        self.embed_item_queue.clear()
        self.ssr_queue = {}

    def post_process(self, content: str):
        """Given the HTML content, post process it to embed assets and render server side rendered components

        This is typically called by :class:`~alliance_platform.frontend.bundler.middleware.BundlerAssetContextMiddleware`, but can
        be called directly if rendering templates of other purposes such as emails of PDFs. ``content`` should be the
        full HTML content of the page, including the ``<head>`` and ``<body>`` tags. This function will handle embedding
        any necessary assets based on calls to :meth:`~alliance_platform.frontend.bundler.context.BundlerAssetContext.queue_embed_file`
        as well as any server side rendered components queued with :meth:`~alliance_platform.frontend.bundler.context.BundlerAssetContext.queue_ssr`.

        This should be the last step before the HTML is outputted to the client.
        """
        try:
            if self.ssr_queue:
                from .ssr import BundlerAssetServerSideRenderer

                global_context = {}
                if self.request:
                    global_context["currentUrl"] = self.request.build_absolute_uri()
                    if resolver := get_ssr_global_context_resolver():
                        global_context.update(resolver(self.request))
                content = BundlerAssetServerSideRenderer(self.ssr_queue).process(content, global_context)
            if not self.embed_item_queue.is_empty():
                if not self.embed_collected_assets_tag_placeholder:
                    warnings.warn(
                        "There are some assets that require embedding but no {% bundler_embed_collected_assets %} exists in the current template."
                    )
                else:
                    tags = []
                    for item in self.embed_item_queue.items:
                        code = item.generate_code(self.html_target)
                        if code:
                            tags.append(code)
                    content = content.replace(self.embed_collected_assets_tag_placeholder, "\n".join(tags))
            elif self.embed_collected_assets_tag_placeholder:
                # no assets to embed but we need to remove the placeholder
                content = content.replace(self.embed_collected_assets_tag_placeholder, "")
            return content
        finally:
            self.embed_item_queue.clear()
            self.ssr_queue = {}
